"""Top-level Stix Model Class."""
# standard library
import itertools
from typing import Union

# third-party
import jmespath

# first-party
from tcex.logger import Logger


class StixModel:
    """STIX base model object."""

    def __init__(self, logger: Logger):
        """Initialize Class properties."""
        self.logger = logger
        self._as_object = None
        self._ipv4 = None
        self._ipv6 = None
        self._registry_key = None
        self._url = None
        self._domain_name = None
        self._email_address = None
        self._indicator = None
        self._relationship = None
        self._indicator_type_details = None

        self._visitors = []

    def _construct_summary(self, data, indicator_type):
        if 'summary' in data.keys():
            return data.get('summary')

        details = self.indicator_type_details.get(indicator_type)

        summary = []
        for field in details.get('fields', []):
            if field in data:
                summary.append(data.get(field))
            else:
                summary.append('')

        return ' : '.join(summary)

    @property
    def indicator_type_details(self):
        """Details for individual ThreatConnect indicator types."""
        if not self._indicator_type_details:
            # first-party
            from tcex.stix.indicator.stix_pattern_helpers import (
                address_stix_pattern_producer,
                asn_stix_pattern_producer,
                cidr_stix_pattern_producer,
                email_address_stix_pattern_producer,
                file_stix_pattern_producer,
                host_stix_pattern_producer,
                registery_key_stix_pattern_producer,
                url_stix_pattern_producer,
            )

            self._indicator_type_details = {
                'host': {
                    'lambda': host_stix_pattern_producer,
                    'api_branch': 'host',
                    'fields': ['text'],
                },
                'url': {
                    'lambda': url_stix_pattern_producer,
                    'api_branch': 'url',
                    'fields': ['text'],
                },
                'emailaddress': {
                    'lambda': email_address_stix_pattern_producer,
                    'api_branch': 'emailaddress',
                    'fields': ['addresses'],
                },
                'asn': {
                    'lambda': asn_stix_pattern_producer,
                    'api_branch': 'asns',
                    'fields': ['as_number'],
                },
                'address': {
                    'lambda': address_stix_pattern_producer,
                    'api_branch': 'address',
                    'fields': ['ip'],
                },
                'cidr': {
                    'lambda': cidr_stix_pattern_producer,
                    'api_branch': 'cidr',
                    'fields': ['ip'],
                },
                'file': {
                    'lambda': file_stix_pattern_producer,
                    'api_branch': 'file',
                    'fields': ['md5', 'sha1', 'sha256'],
                },
                'registry key': {
                    'lambda': registery_key_stix_pattern_producer,
                    'api_branch': 'registryKey',
                    'fields': ['Key Name', 'Value Name', 'Value Type'],
                },
            }
        return self._indicator_type_details

    @property
    def as_object(self):
        """ASN Parser."""
        if not self._as_object:
            from .observables.autonomous_system import StixASObject

            self._as_object = StixASObject(self.logger)
        return self._as_object

    @property
    def ipv4(self):
        """IPv4 Parser."""
        if not self._ipv4:
            from .observables.ip_addr import StixIPv4Object

            self._ipv4 = StixIPv4Object(self.logger)
        return self._ipv4

    @property
    def ipv6(self):
        """IPv6 Parser."""
        if not self._ipv6:
            from .observables.ip_addr import StixIPv6Object

            self._ipv6 = StixIPv6Object(self.logger)
        return self._ipv6

    @property
    def registry_key(self):
        """Registry Key Parser."""
        if not self._registry_key:
            from .observables.registry_key import StixRegistryKeyObject

            self._registry_key = StixRegistryKeyObject(self.logger)
        return self._registry_key

    @property
    def url(self):
        """URL Parser."""
        if not self._url:
            from .observables.url import StixURLObject

            self._url = StixURLObject(self.logger)
        return self._url

    @property
    def email_address(self):
        """Email Address Parser."""
        if not self._email_address:
            from .observables.email_address import StixEmailAddressObject

            self._email_address = StixEmailAddressObject(self.logger)
        return self._email_address

    @property
    def domain_name(self):
        """Domain Name Parser."""
        if not self._domain_name:
            from .observables.domain_name import StixDomainNameObject

            self._domain_name = StixDomainNameObject(self.logger)
        return self._domain_name

    @property
    def indicator(self):
        """Return Indicator Parser."""
        if not self._indicator:
            from .indicator.indicator import StixIndicator

            self._indicator = StixIndicator(self.logger)
        return self._indicator

    @property
    def relationship(self):
        """Relationship Parser."""
        if not self._relationship:
            from .relationship.relationship import Relationship

            self._relationship = Relationship()
        return self._relationship

    def produce(self, tc_data: Union[list, dict], **kwargs):  # pylint: disable=unused-argument
        """Convert ThreatConnect data (in parsed JSON format) into STIX objects.

        Args:
            tc_data: one or more ThreatConnect object dictionaries

        Yields:
            STIX objects
        """
        if not isinstance(tc_data, list):
            tc_data = [tc_data]

        for data in tc_data:
            for indicator_type, normalized_data in self._normalize_tc_objects(data):
                yield from self.relationship.produce(normalized_data)
                yield from self.indicator.produce(normalized_data, indicator_type=indicator_type)

    def _normalize_tc_objects(self, tc_data):
        indicator_type = None
        api_branch = 'indicator'
        data = tc_data.get('data', tc_data)
        for key, values in self.indicator_type_details.items():
            if values.get('api_branch').upper() in [s.upper() for s in data.keys()] or values.get(
                'type', ''
            ).upper() in [s.upper() for s in data.keys()]:
                api_branch = values.get('api_branch')
                indicator_type = key
                break

        if indicator_type is None:
            for key, values in self.indicator_type_details.items():
                for field in values.get('fields'):
                    if field.lower() in map(str.lower, data.keys()):
                        indicator_type = key
                        break
                if indicator_type:
                    break

        tc_data = data.get(api_branch, data)

        if not isinstance(tc_data, list):
            tc_data = [tc_data]

        for data in tc_data:
            _type = indicator_type or data.get('type').lower()
            self.logger.log.error(f'_type: {_type}')
            data['summary'] = self._construct_summary(data, _type)
            for field in self.indicator_type_details.get(_type).get('fields'):
                data.pop(field, '')
            yield _type, data

    def consume(self, stix_data: Union[list, dict]):
        """Convert stix_data (in parsed JSON format) into ThreatConnect objects.

        Args:
            stix_data: one or more stix_data dictionaries

        Yields:
            ThreatConnect objects
        """
        type_mapping = {
            'autonomous-system': self.as_object,
            'domain-name': self.domain_name,
            'email-addr': self.email_address,
            'ipv4-addr': self.ipv4,
            'ipv6-addr': self.ipv6,
            'windows-registry-key': self.registry_key,
            'url': self.url,
            'indicator': self.indicator,
        }

        visitor_mapping = {'relationship': self.relationship}

        if not isinstance(stix_data, list):
            stix_data = [stix_data]

        tc_data = []
        for data in stix_data:
            # Handle a bundle OR just one or more stix objects.
            stix_objects = data.get('objects') if data.get('type') == 'bundle' else data

            for stix_object in stix_objects:
                _type = stix_object.get('type').lower()
                if _type in visitor_mapping:
                    for visitor in visitor_mapping.get(_type).consume(stix_object):
                        self.register_visitor(visitor)
                else:
                    if _type in type_mapping:
                        # sub-parsers return generators, so chain them all together to flatten.
                        tc_data = itertools.chain(
                            tc_data, type_mapping.get(_type).consume(stix_object)
                        )
                    else:
                        # TODO handle unknown stix type
                        pass

        for visitor in self._visitors:
            tc_data = visitor.visit(tc_data)

        yield from tc_data

    def register_visitor(self, visitor):
        """Register a visitor that will be passed all parsed data after consume is through."""
        self._visitors.append(visitor)

    def _map(self, data: Union[list, dict], mapping: dict):

        if isinstance(data, dict):
            data = [data]
        try:
            for d in data:
                mapped_obj = mapping.copy()
                for key, value in mapping.items():
                    if isinstance(value, list):
                        new_list = []
                        for item in value:
                            new_list.append(list(self._map(d, item))[0])

                        mapped_obj[key] = new_list
                    elif isinstance(value, dict):
                        mapped_obj[key] = list(self._map(d, mapped_obj[key]))[0]
                    else:
                        if not value.startswith('@'):
                            mapped_obj[key] = value
                        else:
                            mapped_obj[key] = jmespath.search(f'{value}', jmespath.search('@', d))
                yield mapped_obj
        except Exception:  # pylint: disable=bare-except
            self.logger.log.error(f'Could not map {data} using {mapping}')



"""Request Key File Handler Class"""
# standard library
import logging
import os
import re
import threading
from typing import Optional


class PatternFileHandler(logging.FileHandler):
    """Logger handler for ThreatConnect Exchange Pattern logging.

    This logging handler manages file limits based on a logging pattern. Given
    the pattern and max log count the handler manage the number of log files
    on disk.
    """

    def __init__(
        self,
        filename: str,
        pattern: str,
        mode: Optional[str] = 'a',
        encoding: Optional[str] = None,
        delay: Optional[bool] = False,
        max_log_count: Optional[int] = 100,
    ):
        """Add logic to create log directory if it does not exists.

        Args:
            filename: The name of the logfile.
            pattern: The pattern used to match the log files.
            mode: The write mode for the file.
            encoding: The log file encoding.
            delay: If True, then file opening is deferred until the first call to emit().
            max_log_count: The maximum number of log files to preserve.
        """
        if encoding is None and os.getenv('LANG') is None:
            encoding = 'UTF-8'

        if not os.path.exists(os.path.dirname(filename)):  # pragma: no cover
            os.makedirs(os.path.dirname(filename), exist_ok=True)

        # create pattern
        self.pattern = re.compile(pattern, re.I)

        # trim the number of logfiles to the max log count
        self._trim_log_files(filename, max_log_count)

        super().__init__(filename=filename, mode=mode, encoding=encoding, delay=delay)

    def _trim_log_files(self, filename: str, max_log_count: int) -> None:
        """Trim log files removing the oldest files.

        Args:
            filename: The logfile name.
            max_log_count: The max number of log files to keep.

        """
        log_files = []
        for entry in os.scandir(os.path.dirname(filename)):
            if not entry.is_file():
                continue

            # find only matches
            logfile_name = os.path.basename(entry.path)
            if re.match(self.pattern, logfile_name):
                log_files.append(entry)

        # sort log file entries based on modified time
        log_files.sort(key=lambda e: e.stat().st_mtime, reverse=True)

        # remove oldest logfiles
        for entry in log_files[max_log_count:]:
            try:
                os.remove(entry.path)
            except Exception:  # nosec
                # best effort
                pass

    def emit(self, record: object) -> None:
        """Emit a record.

        Emit logging events only if handler_key matches thread_key.

        Args:
            record: The record to be logged.
        """
        # handler_key and thread_key are added in logger.add_thread_file_handler() method
        # pylint: disable=no-member
        if hasattr(threading.current_thread(), self.thread_key):
            if self.handler_key == getattr(threading.current_thread(), self.thread_key):
                logging.FileHandler.emit(self, record)



"""TcEx testing profile Class."""
# standard library
import json
import re
import sys

# third-party
import colorama as c
import jmespath

# autoreset colorama
c.init(autoreset=True, strip=False)


class Populate:
    """Populate profile env vars and other items."""

    def __init__(self, profile):
        """Initialize Class properties."""
        self.profile = profile

    def replace_env_variables(self, profile_data):
        """Replace any env vars.

        Environment variables replaced follow the pattern ${type:name[=default]}
        e.g. ${env:FOO} or ${env:FOO=bla} to have a default value of "bla" if FOO
        is unset.

        A warning is raised when a substitution is called for, but the source
        does not contain a matching key.  A default value of '' will be used
        when this happens.

        An error is raised if the substitution would result in a new key
        pattern being formed, i.e. if the substitution text contains '${',
        which may or may not be expanded, or break profile expansion.

        Args:
            profile_data (dict): The profile data dict.

        Returns:
            dict: The updated dict.
        """
        profile = json.dumps(profile_data)

        for m in re.finditer(r'\${(env|envs|local|remote):(.*?)}', profile):
            try:
                full_match = m.group(0)
                env_type = m.group(1)  # currently env, envs, local, or remote
                env_key = m.group(2)
                if '=' in env_key:
                    env_key, default_value = env_key.split('=', 1)
                else:
                    default_value = None

                env_value = self.profile.env_store.getenv(env_key, env_type, default_value)
                if env_value is not None:
                    if '${' in env_value:
                        self.profile.log.error(
                            f'Profile replacement value for {full_match} includes '
                            f'recursive expansion {env_value}'
                        )
                    profile = profile.replace(full_match, env_value)
            except IndexError:
                print(f'{c.Fore.YELLOW}Could not replace variable {full_match}).')
        return json.loads(profile)

    def replace_tc_variables(self, profile_data):
        """Replace all of the TC output variables in the profile with their correct value.

        Args:
            profile_data (dict): The profile data dict.

        Returns:
            dict: The updated dict.
        """
        if 'jmespath' not in sys.modules:
            print(
                f'{c.Fore.RED}Missing jmespath module. Try '
                f'installing "pip install tcex[development]"'
            )
            sys.exit(1)

        profile = json.dumps(profile_data)

        for data in self.profile.tc_staged_data:
            key = data.get('key')
            data_value = data.get('data')

            for m in re.finditer(r'\${tcenv:' + str(key) + r':(.*?)}', profile):
                try:
                    full_match = m.group(0)
                    jmespath_expression = m.group(1)

                    if jmespath_expression:
                        value = jmespath.search(jmespath_expression, data_value)
                        profile = profile.replace(full_match, str(value))
                except IndexError:
                    print(f'{c.Fore.YELLOW}Invalid variable found {full_match}.')
        return json.loads(profile)



"""ThreatConnect TI Host"""
# standard library
from urllib.parse import quote_plus

from ..indicator import Indicator


class Host(Indicator):
    """Unique API calls for Host API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties.

        Args:
            hostname (str): The value for this Indicator.
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): [Read-Only] The date timestamp the Indicator was created.
            last_modified (str, kwargs): [Read-Only] The date timestamp the Indicator was last
                modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            dns_active (bool, kwargs): If True DNS active is enabled for this indicator.
            whois_active (bool, kwargs): If True WhoIs active is enabled for this indicator.
        """
        super().__init__(tcex, sub_type='Host', api_entity='host', api_branch='hosts', **kwargs)
        self.unique_id = kwargs.get('unique_id', self._data.get('hostName'))
        self._data['hostName'] = self.unique_id
        if self.unique_id:
            self.unique_id = quote_plus(self.fully_decode_uri(self.unique_id))

    def can_create(self):
        """Return True if file can be create.

        If the hostName has been provided returns that the File can be created, otherwise
        returns that the Host cannot be created.
        """
        return not self.data.get('hostName') is None

    def _set_unique_id(self, json_response):
        """Set the unique_id provided a json response.

        Args:
            json_response:
        """
        self.unique_id = quote_plus(self.fully_decode_uri(json_response.get('hostName', '')))

    def dns_resolution(self):
        """Update the Host DNS resolution

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.dns_resolution(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner
        )

    def set_dns_resolution(self, value):
        """Update the Host DNS resolution

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.set_dns_resolution(
            self.api_type, self.api_branch, self.unique_id, value, owner=self.owner
        )

    def set_whois(self, value):
        """Update the Indicators WhoIs

        Args:
            value:
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])
        return self.tc_requests.set_whois(
            self.api_type, self.api_branch, self.unique_id, value, owner=self.owner
        )



"""ThreatConnect TI Adversary """
# standard library
from urllib.parse import quote_plus

from .mappings import Mappings


class Task(Mappings):
    """Unique API calls for Tasks API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties.

        Valid status:
        + Not Started
        + In Progress
        + Completed
        + Waiting on Someone
        + Deferred

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
            status (str, kwargs): Not started, In Progress, Completed, Waiting on Someone, Deferred
            due_date (str, kwargs): Converted to %Y-%m-%dT%H:%M:%SZ date format
            reminder_date (str, kwargs): Converted to %Y-%m-%dT%H:%M:%SZ date format
            escalation_date (str, kwargs): Converted to %Y-%m-%dT%H:%M:%SZ date format
        """

        super().__init__(
            tcex,
            main_type='Task',
            api_type='tasks',
            sub_type=None,
            api_entity='task',
            api_branch=None,
            owner=kwargs.pop('owner', None),
        )
        for arg, value in kwargs.items():
            self.add_key_value(arg, value)

    def _set_unique_id(self, json_response):
        """Set the unique id of the Group."""
        self.unique_id = json_response.get('id', '')

    @property
    def _metadata_map(self):
        """Return metadata map for Group objects."""
        return {
            'due_date': 'dueDate',
            'reminder_date': 'reminderDate',
            'escalation_date': 'escalationDate',
        }

    @property
    def as_entity(self):
        """Return the entity representation of the Task."""
        return {
            'type': 'Task',
            'value': self.name,
            'id': int(self.unique_id) if self.unique_id else None,
        }

    @property
    def name(self):
        """Return Task name."""
        return self._data.get('name')

    @name.setter
    def name(self, name):
        """Set the Group name."""
        self._data['name'] = name

    @staticmethod
    def is_task():
        """Return True if object is a task."""
        return True

    def add_assignee(self, assignee):
        """Add the desired assignee from the Task.

        Args:
            assignee (str): The assignee username

        Return:
            obj: The response of the POST.

        """
        return self.assignee(assignee)

    def add_key_value(self, key, value):
        """Convert the value and adds it as a data field.

        Args:
            key:
            value:
        """
        key = self._metadata_map.get(key, key)
        if key in ['unique_id', 'id']:
            self._unique_id = quote_plus(str(value))
        elif key in ['dueDate', 'reminderDate', 'escalationDate']:
            self._data[key] = self._utils.datetime.format_datetime(
                value, date_format='%Y-%m-%dT%H:%M:%SZ'
            )
        else:
            self._data[key] = value

    def add_escalatee(self, escalatee):
        """Add the desired escalatee from the Task.

        Args:
            escalatee (str): The escalatee username

        Return:
            obj: The response of the POST.

        """
        return self.escalatee(escalatee)

    def assignee(self, assignee, action='ADD'):
        """General method to perform actions on assignees

        Valid Actions:
        + ADD
        + GET
        + DELETE

        Args:
            assignee (str): The username of the assignee.
            action: [ADD, DELETE, GET] the action to be done on the escalatee. Defaults to
                ADD if not provided.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.assignee(self.api_type, self.unique_id, assignee, action=action)

    def assignees(self):
        """Yield the Assignee Users"""
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.assignees(self.api_type, self.unique_id)

    def can_create(self):
        """Return True if the Object can be created.

        Return:
            bool: Boolean value indicating whether the object can be created.
        """
        if self.data.get('name'):
            return True
        return False

    def delete_assignee(self, assignee):
        """Delete the desired assignee from the Task.

        Args:
            assignee (str): The assignee username

        Return:
            obj: The response of the DELETE.

        """
        return self.assignee(assignee, action='DELETE')

    def delete_escalatee(self, escalatee):
        """Delete the desired escalatee from the Task.

        Args:
            escalatee (str): The escalatee username

        Return:
            obj: The response of the DELETE.

        """
        return self.escalatee(escalatee, action='DELETE')

    def due_date(self, due_date):
        """Update the task due_date

        Args:
            due_date: Converted to %Y-%m-%dT%H:%M:%SZ date format
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        due_date = self._utils.datetime.format_datetime(due_date, date_format='%Y-%m-%dT%H:%M:%SZ')
        self._data['dueDate'] = due_date
        request = {'dueDate': due_date}
        return self.tc_requests.update(self.api_type, self.api_sub_type, self.unique_id, request)

    def escalatee(self, escalatee, action='ADD'):
        """General method to perform actions on escalatees

        Valid Actions:
        + ADD
        + GET
        + DELETE

        Args:
            escalatee (str): The username of the escalatee.
            action: [ADD, DELETE, GET] the action to be done on the escalatee. Defaults to
                ADD if not provided.

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.escalatee(self.api_type, self.unique_id, escalatee, action=action)

    def escalatees(self):
        """Yield the Escalatees Users"""
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.escalatees(self.api_type, self.unique_id)

    def escalation_date(self, escalation_date):
        """Update the task escalation_date

        Args:
            escalation_date: Converted to %Y-%m-%dT%H:%M:%SZ date format
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        escalation_date = self._utils.datetime.format_datetime(
            escalation_date, date_format='%Y-%m-%dT%H:%M:%SZ'
        )
        self._data['escalationDate'] = escalation_date
        request = {'escalationDate': escalation_date}
        return self.tc_requests.update(self.api_type, self.api_sub_type, self.unique_id, request)

    def get_assignee(self, assignee):
        """Retrieve the desired assignee from the Task.

        Args:
            assignee (str): The assignee username

        Return:
            obj: The response of the GET.

        """
        return self.assignee(assignee, action='GET')

    def get_escalatee(self, escalatee):
        """Retrieve the desired escalatee from the Task.

        Args:
            escalatee (str): The escalatee username

        Return:
            obj: The response of the GET.

        """
        return self.escalatee(escalatee, action='GET')

    def reminder_date(self, reminder_date):
        """Update the task reminder_date

        Args:
            reminder_date: Converted to %Y-%m-%dT%H:%M:%SZ date format
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        reminder_date = self._utils.datetime.format_datetime(
            reminder_date, date_format='%Y-%m-%dT%H:%M:%SZ'
        )
        self._data['reminderDate'] = reminder_date
        request = {'reminderDate': reminder_date}
        return self.tc_requests.update(self.api_type, self.api_sub_type, self.unique_id, request)

    def status(self, status):
        """Update the Task Status

        Valid status:
        + Not Started
        + In Progress
        + Completed
        + Waiting on Someone
        + Deferred

        Args:
            status: Not Started, In Progress, Completed, Waiting on Someone, Deferred
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        self._data['status'] = status
        request = {'status': status}
        return self.tc_requests.update(self.api_type, self.api_sub_type, self.unique_id, request)



"""Visitor producer for STIX Relationship Object.

import stix2
from tcex.batch import Batch

see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070673
"""
# standard library
from typing import Iterable, List, Union

# third-party
import stix2

# first-party
from tcex.batch import Batch
from tcex.stix.visitor import Visitor, VisitorProducer


class RelationshipVisitor(Visitor):
    """Visitor that creates associations between TC objects based on a STIX Relationship."""

    def __init__(self, target_ref: str, source_ref: str):
        """Visitor that creates associations between TC objects based on a STIX Relationship.

        Args:
            target_ref: xid of TC object on one side of the association.
            source_ref: xid of the TC object on the other side of the association.
        """
        super().__init__()
        self.target_ref = target_ref
        self.source_ref = source_ref

        self.target = None
        self.source = None

        self.done = False

    def visit(self, data: Iterable[dict]) -> Iterable[dict]:
        """Look for the parsed objects referred to by target_ref and source_ref and associate them.

         Associations an only be group->group or indicator->group.  Indicator-indicator associations
         are not supported.

         Args:
            data: generator of parsed data.

        Yields:
            parsed data with associations added.
        """
        for d in data:
            if self.done:
                yield d  # get out asap if we're done to reduce overhead.
            else:
                if d.get('xid') == self.target_ref:
                    self.target = d
                if d.get('xid') == self.source_ref:
                    self.source = d

                if self.target and self.source:  # if we've already found our targets...
                    if self._is_group(self.target) and self._is_group(self.source):
                        # both are groups:
                        self.target.setdefault('associatedGroupXid', []).append(
                            self.source.get('xid')
                        )
                        self.source.setdefault('associatedGroupXid', []).append(
                            self.target.get('xid')
                        )
                    elif self._is_group(self.target) and not self._is_group(self.source):
                        # source is an indicator, target is a group
                        self.source.setdefault('associatedGroups', []).append(
                            {'groupXid': self.target.get('xid')}
                        )
                    elif self._is_group(self.source) and not self._is_group(self.target):
                        # target is an indicator, source is a group
                        self.target.setdefault('associatedGroups', []).append(
                            {'groupXid': self.source.get('xid')}
                        )
                    else:
                        # TODO handle indicator-to-indicator, we can't do that in TC
                        pass

                    self.done = True

                    yield self.target
                    yield self.source
                else:
                    yield d

    @staticmethod
    def _is_group(d):
        return d.get('name') is not None


class Relationship(VisitorProducer):
    """Visitor producer for STIX Relationship Object.

    see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070673

    Args:
        model: StixModel instance to register visitor on.
    """

    def consume(self, stix_data: Union[List[dict], dict]) -> Iterable[Visitor]:
        """Register a visitor to create ThreatConnect associations for STIX relationship.

        Args:
            stix_data: One or more STIX Relationship objects.

        Yields:
            Visitors that will create ThreatConnect Associations based on a STIX Relationship.
        """
        if not isinstance(stix_data, list):
            stix_data = [stix_data]

        for data in stix_data:
            yield RelationshipVisitor(data.get('target_ref'), data.get('source_ref'))

    def produce(self, tc_data: Union[list, dict]) -> Iterable[Visitor]:
        """Yield visitors that will apply relationship objects to a Stream of STIX data.

        Args:
            tc_data: the ThreatConnect data to produce Relationships based on.

        Yields:
            Visitors that will create STIX Relationship objects.
        """
        if not isinstance(tc_data, list):
            tc_data = [tc_data]

        for data in tc_data:
            associations = data.pop('associations', [])
            for association in associations:
                source_xid = Batch.generate_xid([data.get('summary'), data.get('ownerName')])
                target_xid = Batch.generate_xid(
                    [association.get('summary'), association.get('ownerName')]
                )

                yield stix2.Relationship(
                    source_ref=f'indicator--{source_xid}',
                    target_ref=f'indicator--{target_xid}',
                    relationship_type='uses',
                )



"""TcEx testing Framework."""
# standard library
import logging
import math
import os

from ..logger import RotatingFileHandlerCustom
from ..logger.trace_logger import TraceLogger


class TestLogger(TraceLogger):
    """Custom logger for test Case"""

    def data(self, stage, label, data, level='info'):
        """Log validation data in a specific format for test case.

        Args:
            stage (str): The stage of the test phase (e.g., setup, staging, run, validation, etc)
            label (str): The label for the provided data.
            data (str): The data to log.
            level (str, optional): The logging level to write the event. Defaults to 'info'.
        """
        stage = f'[{stage}]'
        stage_width = 25 - len(level)
        msg = f'{stage!s:>{stage_width}} : {label!s:<20}: {data!s:<50}'
        level = logging.getLevelName(level.upper())
        self.log(level, msg)

    def title(self, title, separator='-'):
        """Log validation data.

        Args:
            title (str): The title for the section of data to be written using data method.
            separator (str, optional): The value to use as a separator. Defaults to '-'.
        """
        separator = separator * math.ceil((100 - len(str(title))) / 2)
        self.log(logging.INFO, f'{separator} {title} {separator}')


# create logger based on custom TestLogger
logging.setLoggerClass(TestLogger)

# init logger
logger = logging.getLogger('TestCase')

# set logger level
logger.setLevel(logging.TRACE)

# create rotation filehandler
lfh = RotatingFileHandlerCustom(filename='log/tests.log')

# get logging level from OS env or default to debug
logging_level = logging.getLevelName(os.getenv('TCEX_TEST_LOGGING_LEVEL', 'debug').upper())

# set handler logging level
lfh.setLevel(logging_level)

# create formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
if logging_level < 10:
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        '(%(filename)s:%(funcName)s:%(lineno)d:%(threadName)s)'
    )

# set formatter
lfh.setFormatter(formatter)

# add handler
logger.addHandler(lfh)



"""ThreatConnect Case Management Collection"""
# third-party
from requests.exceptions import ProxyError

from .tql import TQL


class CommonCaseManagementCollection:
    """Common Class for Case Management Collection

    Object encapsulates common methods used by child classes.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # Example of params input
        {
            'result_limit': 100,  # How many results are retrieved.
            'result_start': 10,  # Starting point on retrieved results.
            'fields': ['caseId', 'summary']  # Additional fields returned on the results
        }

    Args:
        tcex ([type]): [description]
        api_endpoint ([type]): [description]
        tql_filters ([type], optional): [description]. Defaults to None.
        initial_response ([type], optional): [description]. Defaults to None.
        params(dict, optional): Dict of the params to be sent while
                    retrieving the Case Management objects.

    Returns:
        [type]: [description]

    Yields:
        [type]: [description]
    """

    def __init__(
        self, tcex, api_endpoint, tql_filters=None, initial_response=None, params=None,
    ):
        """Initialize Class properties."""

        self._added_items = []
        self._initial_response = initial_response
        self._params = params or {}
        self._tql_data = None
        self._tql_filters = tql_filters or []
        self.api_endpoint = api_endpoint.value
        self.tcex = tcex
        self.tql = TQL()

    def __len__(self):
        """Return the length of the collection."""
        parameters = self.params
        parameters['result_limit'] = 1
        tql_string = self.tql.raw_tql
        if not self.tql.raw_tql:
            self.tql.filters = self._tql_filters + self.tql.filters
            tql_string = self.tql.as_str
        if tql_string:
            parameters['tql'] = tql_string

        # convert all keys to camel case
        for k, v in list(parameters.items()):
            del parameters[k]
            k = self.tcex.utils.snake_to_camel(k)
            parameters[k] = v

        r = self.tcex.session.get(self.api_endpoint, params=parameters)
        return r.json().get('count')

    def __str__(self):
        """Object iterator"""
        printable_string = ''
        for obj in self.iterate(initial_response=self.initial_response):
            printable_string += f'{"-" * 50}\n'
            printable_string += str(obj)
        return printable_string

    @property
    def _filter_map_method(self):
        """Return a property for keywork mapping."""
        method_data = (
            f'\n{" " * 4}@property\n'
            f'\n{" " * 4}def keyword_map(self):\n'
            f'{" " * 8}"""Return keyword to method name map."""\n\n'
            f'{" " * 8}return {{\n'
        )
        for t in sorted(self.tql_data, key=lambda i: i['keyword']):
            method_data += f"""{" " * 12}'{t.get('keyword')}': '{t.get('keyword')}',\n"""
        method_data += f'{" " * 8}}}\n'

        return method_data

    @property
    def _filter_class(self):
        """Return a Filter Class current object."""
        filter_class = (
            f'\nclass Filter{self.__class__.__name__}(Filter):\n'
            f'{" " * 4}"""Filter Object for {self.__class__.__name__}"""\n'
        )

        for t in sorted(self.tql_data, key=lambda i: i['keyword']):
            class_name = self.tcex.utils.camel_to_space(self.__class__.__name__).title()
            description = t.get('description')
            keyword = t.get('keyword')
            keyword_snake = self.tcex.utils.camel_to_snake(keyword)

            # get the arg type
            keyword_type = 'str'
            tql_type = 'TQL.Type.STRING'
            if t.get('type') in ['Integer', 'Long']:
                keyword_type = 'int'
                tql_type = 'TQL.Type.INTEGER'
            elif t.get('type') == 'Boolean':
                keyword_type = 'bool'
                tql_type = 'TQL.Type.BOOLEAN'

            comment = ''
            if keyword_snake in ['id', 'type']:
                comment = '  # pylint: disable=redefined-builtin'

            if keyword_snake == 'has_artifact':
                filter_class += (
                    f'\n{" " * 4}@property\n'
                    f'{" " * 4}def has_artifact(self):\n'
                    f'{" " * 8}"""Return **FilterArtifacts** for further filtering."""\n'
                    f'{" " * 8}from .artifact import FilterArtifacts\n\n'
                    f'{" " * 8}artifacts = FilterArtifacts'
                    f'(ApiEndpoints.ARTIFACTS, self._tcex, TQL())\n'
                    f"""{" " * 8}self._tql.add_filter('hasArtifact', """
                    f'TQL.Operator.EQ, artifacts, TQL.Type.SUB_QUERY)\n'
                    f'{" " * 8}return artifacts\n'
                )
            elif keyword_snake == 'has_case':
                filter_class += (
                    f'\n{" " * 4}@property\n'
                    f'{" " * 4}def has_case(self):\n'
                    f'{" " * 8}"""Return **FilterCases** for further filtering."""\n'
                    f'{" " * 8}from .case import FilterCases  # pylint: disable=cyclic-import\n\n'
                    f'{" " * 8}cases = FilterCases(ApiEndpoints.CASES, self._tcex, TQL())\n'
                    f"""{" " * 8}self._tql.add_filter('hasCase', """
                    f'TQL.Operator.EQ, cases, TQL.Type.SUB_QUERY)\n'
                    f'{" " * 8}return cases\n'
                )
            elif keyword_snake == 'has_note':
                filter_class += (
                    f'\n{" " * 4}@property\n'
                    f'{" " * 4}def has_note(self):\n'
                    f'{" " * 8}"""Return **FilterNotes** for further filtering."""\n'
                    f'{" " * 8}from .note import FilterNotes\n\n'
                    f'{" " * 8}notes = FilterNotes(ApiEndpoints.NOTES, self._tcex, TQL())\n'
                    f"""{" " * 8}self._tql.add_filter('hasNote', """
                    f'TQL.Operator.EQ, notes, TQL.Type.SUB_QUERY)\n'
                    f'{" " * 8}return notes\n'
                )
            elif keyword_snake == 'has_tag':
                filter_class += (
                    f'\n{" " * 4}@property\n'
                    f'{" " * 4}def has_tag(self):\n'
                    f'{" " * 8}"""Return **FilterTags** for further filtering."""\n'
                    f'{" " * 8}from .tag import FilterTags\n\n'
                    f'{" " * 8}tags = FilterTags(ApiEndpoints.TAGS, self._tcex, TQL())\n'
                    f"""{" " * 8}self._tql.add_filter('hasTag', """
                    f'TQL.Operator.EQ, tags, TQL.Type.SUB_QUERY)\n'
                    f'{" " * 8}return tags\n'
                )
            elif keyword_snake == 'has_task':
                filter_class += (
                    f'\n{" " * 4}@property\n'
                    f'{" " * 4}def has_task(self):\n'
                    f'{" " * 8}"""Return **FilterTask** for further filtering."""\n'
                    f'{" " * 8}from .task import FilterTasks\n'
                    f'{" " * 8}tasks = FilterTasks(ApiEndpoints.TASKS, self._tcex, TQL())\n'
                    f"""{" " * 8}self._tql.add_filter('hasTask', """
                    f'TQL.Operator.EQ, tasks, TQL.Type.SUB_QUERY)\n'
                    f'{" " * 8}return tasks\n'
                )
            else:
                # build method
                filter_class += (
                    f'\n{" " * 4}def {keyword_snake}(self, operator, {keyword_snake}):{comment}\n'
                    f'{" " * 8}"""Filter {class_name} based on **{keyword}** keyword.\n\n'
                    f'{" " * 8}Args:\n'
                    f'{" " * 12}operator (enum): The operator enum for the filter.\n'
                    f'{" " * 12}{keyword_snake} ({keyword_type}): {description}.\n'
                    f'{" " * 8}"""\n'
                    f"{' ' * 8}self._tql.add_filter('{keyword}', operator, "
                    f'{keyword_snake}, {tql_type})\n'
                )

        return filter_class

    @property
    def added_items(self):
        """Return the added items to the collection"""
        return self._added_items

    @added_items.setter
    def added_items(self, added_items):
        """Set the added items to the collection"""
        self._added_items = added_items

    def entity_map(self, entity):  # pragma: no cover
        """Stub for common method."""
        raise NotImplementedError('Child class must implement this method.')

    @property
    def filter(self):  # pragma: no cover
        """Return filter method."""
        raise NotImplementedError('Child class must implement this method.')

    @property
    def initial_response(self):
        """Return the initial response of the case management object collection."""
        return self._initial_response

    def iterate(self, initial_response=None):
        """Iterate over the case management object collection objects.

        Args:
            initial_response ([type], optional): [description]. Defaults to None.

        Yields:
            [type]: [description]
        """
        parameters = self.params

        # convert all keys to camel case
        for k, v in list(parameters.items()):
            del parameters[k]
            k = self.tcex.utils.snake_to_camel(k)
            parameters[k] = v

        url = self.api_endpoint

        tql_string = self.tql.raw_tql
        if not self.tql.raw_tql:
            self.tql.filters = self._tql_filters + self.tql.filters
            tql_string = self.tql.as_str

        if tql_string:
            parameters['tql'] = tql_string

        if initial_response:
            url = initial_response.get('next_url', None)
            entities = initial_response.get('data', [])
            for entity in entities:
                yield self.entity_map(entity)
            if not url:
                return

        while True:
            r = None
            try:
                r = self.tcex.session.get(url, params=parameters)
                self.tcex.log.debug(
                    f'Method: ({r.request.method.upper()}), '
                    f'Status Code: {r.status_code}, '
                    f'URl: ({r.url})'
                )
                self.tcex.log.trace(f'response: {r.text}')
            except (ConnectionError, ProxyError):  # pragma: no cover
                self.tcex.handle_error(
                    951, ['OPTIONS', 407, '{\"message\": \"Connection Error\"}', self.api_endpoint]
                )

            if not self.success(r):
                err = r.text or r.reason
                self.tcex.handle_error(950, [r.status_code, err, r.url])

            # reset some vars
            parameters = {}

            data = r.json().get('data', [])
            url = r.json().pop('next', None)

            for result in data:
                yield self.entity_map(result)

            if not url:
                yield from self.added_items
                break

    @staticmethod
    def list_as_dict(added_items):
        """Return the dict representation of the case management collection object."""
        as_dict = {'data': []}
        for item in added_items:
            as_dict['data'].append(item.as_dict)
        return as_dict

    @property
    def params(self):
        """Return the parameters of the case management object collection."""
        return self._params

    @params.setter
    def params(self, params):
        """Set the parameters of the case management object collection."""
        self._params = params

    @staticmethod
    def success(r):
        """Validate the response is valid.

        Args:
            r (requests.response): The response object.

        Returns:
            bool: True if status is "ok"
        """
        status = True
        if r.ok:
            try:
                if r.json().get('status') != 'Success':  # pragma: no cover
                    status = False
            except Exception:  # pragma: no cover
                status = False
        else:
            status = False
        return status

    @property
    def timeout(self):
        """Return the timeout of the case management object collection."""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        """Set the timeout of the case management object collection."""
        self._timeout = timeout

    @property
    def tql_data(self):
        """Return TQL data keywords."""
        if self._tql_data is None:
            r = self.tcex.session.options(f'{self.api_endpoint}/tql', params={})
            if r.ok:
                self._tql_data = r.json()['data']

        return self._tql_data



"""ThreatConnect Requests Session"""
# standard library
import base64
import hashlib
import hmac
import time
from urllib.parse import quote

# third-party
import urllib3
from requests import Session, adapters, auth
from urllib3.util.retry import Retry

# first-party
from tcex.utils import Utils

# disable ssl warning message
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class HmacAuth(auth.AuthBase):
    """ThreatConnect HMAC Authorization"""

    def __init__(self, access_id, secret_key):
        """Initialize the Class properties."""
        super().__init__()
        self._access_id = access_id
        self._secret_key = secret_key

    def __call__(self, r):
        """Override of parent __call__ method."""
        timestamp = int(time.time())
        signature = f'{r.path_url}:{r.method}:{timestamp}'
        hmac_signature = hmac.new(
            self._secret_key.encode(), signature.encode(), digestmod=hashlib.sha256
        ).digest()
        authorization = f'TC {self._access_id}:{base64.b64encode(hmac_signature).decode()}'
        r.headers['Authorization'] = authorization
        r.headers['Timestamp'] = timestamp
        return r


# APP-102 - adding ability to create sessions with different
# TC credentials for accessing special API endpoints
class TestSession(Session):
    """ThreatConnect REST API Requests Session

    Args:
        api_access_id (str): The ThreatConnect API access id.
        api_secret_key (str): The ThreatConnect API secret key.
        proxy_host (str, kwargs): The proxy hostname.
        proxy_port (int, kwargs): The proxy port number.
        proxy_user (str, kwargs): The proxy username.
        proxy_pass (str, kwargs): The proxy password.
        verify (bool, kwargs): If True SSL verification of the host will be performed.
    """

    def __init__(self, api_access_id, api_secret_key, api_path, logger, **kwargs):
        """Initialize the Class properties."""
        super().__init__()
        self.api_path = api_path
        self.api_access_id = api_access_id
        self.api_secret_key = api_secret_key
        self.log = logger
        self.proxy_host = kwargs.get('proxy_host')
        self.proxy_port = kwargs.get('proxy_port')
        self.proxy_user = kwargs.get('proxy_user')
        self.proxy_pass = kwargs.get('proxy_pass')

        # properties
        self.auth = None
        self.utils = Utils()

        # Update User-Agent
        self.headers.update({'User-Agent': 'TcEx Testing'})

        # Set Proxy
        self.proxies = self.proxy_config

        # Add Retry
        self.retry()

        # Set Verify
        self.verify = kwargs.get('verify', True)

    def _configure_auth(self):
        """Return Auth property for session."""
        if self.api_access_id and self.api_secret_key:
            try:
                # for external Apps or testing Apps locally
                self.auth = HmacAuth(self.api_access_id, self.api_secret_key)
                self.log.debug('Using HMAC authorization.')
            except AttributeError:  # pragma: no cover
                raise RuntimeError('No valid ThreatConnect API credentials provided.')
        else:  # pragma: no cover
            raise RuntimeError('No valid ThreatConnect API credentials provided.')

    @property
    def proxy_config(self):
        """Format the proxy configuration for Python Requests module.

        Generates a dictionary for use with the Python Requests module format
        when proxy is required for remote connections.

        **Example Response**
        ::

            {"http": "http://user:pass@10.10.1.10:3128/"}

        Returns:
           (dictionary): Dictionary of proxy settings
        """
        proxies = {}
        if self.proxy_host is not None and self.proxy_port is not None:

            if self.proxy_user is not None and self.proxy_pass is not None:
                proxy_user = quote(self.proxy_user, safe='~')
                proxy_pass = quote(self.proxy_pass, safe='~')

                # proxy url with auth
                proxy_url = f'{proxy_user}:{proxy_pass}' f'@{self.proxy_host}:{self.proxy_port}'
            else:
                # proxy url without auth
                proxy_url = f'{self.proxy_host}:{self.proxy_port}'
            proxies = {'http': f'http://{proxy_url}', 'https': f'https://{proxy_url}'}
        return proxies

    def request(self, method, url, **kwargs):  # pylint: disable=arguments-differ
        """Override request method disabling verify on token renewal if disabled on session."""
        if self.auth is None:
            self._configure_auth()

        if not url.startswith('https'):
            # automatically add api path to each request
            url = f'{self.api_path}{url}'
        response = super().request(method, url, **kwargs)

        # APP-79 - adding logging of request as curl commands
        self.log.debug(self.utils.requests_to_curl(response.request, verify=self.verify))

        return response

    def retry(self, retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504)):
        """Add retry to Requests Session

        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry
        """
        retries = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        # mount all https requests
        self.mount('https://', adapters.HTTPAdapter(max_retries=retries))



"""Parser for STIX Indicator Objects.

import stix2
see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070633
"""
# standard library
import uuid
from typing import Dict, List, Union

# third-party
import stix2
from dendrol import Pattern
from dendrol.lang.STIXPatternListener import STIXPatternListener
from dendrol.lang.STIXPatternParser import STIXPatternParser

# first-party
from tcex.batch import Batch
from tcex.stix import StixModel  # pylint: disable=cyclic-import


class StixIndicator(StixModel):
    """Parser for STIX Indicator Objects.

    see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070633
    """

    def produce(self, tc_data: Union[list, dict], **kwargs):
        """Produce a STIX Indicator from a ThreatConnect Indicator.

        Args:
            tc_data: ThreatConnect indicator(s)

        Yields:
            A STIX Indicator.
        """
        if not isinstance(tc_data, list):
            tc_data = [tc_data]

        for data in tc_data:
            _type = kwargs.get('indicator_type') or data.get('type')
            indicator_details = self.indicator_type_details.get(_type.lower())
            if not indicator_details:
                continue
            labels = []
            description = None
            latest = None
            for tag in data.get('tag', []):
                labels.append(tag.get('name'))
            for attribute in data.get('attribute', []):
                if attribute.get('type').lower() == 'description':
                    value = attribute.get('value')
                    last_modified = attribute.get('lastModified'), '%y-%m-%d'
                    if attribute.get('displayed'):
                        description = value
                        break
                    if not latest or latest > last_modified:
                        latest = last_modified
                        description = value

            id_ = f'''{data.get('ownerName').lower()}-{_type.lower()}-{data.get('summary')}'''
            self.logger.log.error(f'indicator--{id_}')
            id_ = uuid.uuid5(uuid.NAMESPACE_X500, id_)

            # {
            #     "type": "indicator",
            #     "spec_version": "2.1",
            #     "id": "indicator--78e53a5c-9510-4a95-88b1-dbac1ee60ca5",
            #     "created": "2020-09-08T19:16:25.481964Z",
            #     "modified": "2020-09-08T19:16:25.481964Z",
            #     "name": "TCI - kqvri.com",
            #     "description": "visible description",
            #     "indicator_types": [
            #         "malicious-activity"
            #     ],
            #     "pattern": "[domain-name:value = 'kqvri.com']",
            #     "pattern_type": "stix",
            #     "pattern_version": "2.1",
            #     "valid_from": "2020-09-08T19:16:25.481964Z",
            #     "labels": [
            #         "test1"
            #     ]
            # }

            if data.get('rating', None) == 0:
                labels.append('Threat Rating: Unknown')
            elif data.get('rating', None) == 1:
                labels.append('Threat Rating: Suspicious')
            elif data.get('rating', None) == 2:
                labels.append('Threat Rating: Low')
            elif data.get('rating', None) == 3:
                labels.append('Threat Rating: Moderate')
            elif data.get('rating', None) == 4:
                labels.append('Threat Rating: High')
            elif data.get('rating', None) == 5:
                labels.append('Threat Rating: Very High')

            yield stix2.Indicator(
                confidence=data.get('confidence'),
                labels=labels,
                created=data.get('dateAdded'),
                modified=data.get('lastModified'),
                description=description,
                id=f'indicator--{id_}',
                name=f'{data.get("summary")}',
                lang='en',
                pattern_version='2.1',
                indicator_types=['malicious-activity'],
                pattern_type='stix',
                pattern=indicator_details.get('lambda')(data),
            )

    def consume(self, stix_data: Union[list, dict]):
        """Produce a ThreatConnect object from a STIX 2.0 JSON object.

        Args:
            stix_data: STIX Indicator objects to parse.

        Yields:
            A ThreatConnect Signature and optionally ThreatConnect indicators.
        """
        if not isinstance(stix_data, list):
            stix_data = [stix_data]

        for data in stix_data:
            if data.get('pattern_type', 'stix') != 'stix':
                continue

            signature = {
                'type': 'Signature',
                'xid': data.get('id'),
                'name': data.get('name'),
            }
            if data.get('pattern_type', 'stix') == 'stix':
                # We only parse indicators out of stix patterns...
                pattern = Pattern(data.get('pattern'))
                s = STIXListener()
                pattern.walk(s)

                yield from StixIndicator._default_consume_handler(s.indicators, signature)
                yield from StixIndicator._ip_consume_handler(s.indicators, signature)
                yield from StixIndicator._file_consume_handler(s.indicators, signature)

            yield signature

    @staticmethod
    def _default_consume_handler(stix_indicators: List[Dict[str, str]], signature):
        type_map = {
            'url:value': 'URL',
            'email-addr:value': 'EmailAddress',
            'domain-name:value': 'Host',
            'autonomous-system:name': 'ASN',
        }

        for i in filter(lambda i: i.get('path') in type_map.keys(), stix_indicators):
            path = i.get('path')
            value = i.get('value')
            indicator_type = type_map.get(path)

            yield {
                'type': indicator_type,
                'summary': value,
                'xid': Batch.generate_xid([signature.get('xid'), value]),
                'associatedGroups': [{'groupXid': signature.get('xid')}],
            }

    @staticmethod
    def _ip_consume_handler(stix_indicators: List[Dict[str, str]], signature):
        for i in filter(
            lambda i: i.get('path') in ['ipv4-addr:value', 'ipv6-addr:value'], stix_indicators
        ):
            path = i.get('path')
            value = i.get('value')
            if path == 'ipv4-addr:value':
                if '/' in value and value.split('/')[1] != '32':  # this is a CIDR
                    yield {
                        'type': 'CIDR',
                        'summary': value,
                        'xid': Batch.generate_xid([signature.get('xid'), value]),
                        'associatedGroups': [{'groupXid': signature.get('xid')}],
                    }
                else:  # this is an address
                    yield {
                        'type': 'Address',
                        'summary': value.split('/')[0],
                        'xid': Batch.generate_xid([signature.get('xid'), value]),
                        'associatedGroups': [{'groupXid': signature.get('xid')}],
                    }
            elif path == 'ipv6-addr:value':
                if '/' in value and value.split('/')[1] != '128':  # this is a CIDR
                    yield {
                        'type': 'CIDR',
                        'summary': value,
                        'xid': Batch.generate_xid([signature.get('xid'), value]),
                        'associatedGroups': [{'groupXid': signature.get('xid')}],
                    }
                else:  # this is an address
                    yield {
                        'type': 'Address',
                        'summary': value.split('/')[0],
                        'xid': Batch.generate_xid([signature.get('xid'), value]),
                        'associatedGroups': [{'groupXid': signature.get('xid')}],
                    }

    @staticmethod
    def _file_consume_handler(stix_indicators: List[Dict[str, str]], signature: dict):
        file_indicators = list(filter(lambda i: 'file:hashes' in i.get('path'), stix_indicators))

        sha256_indicators = list(
            filter(lambda i: 'SHA-156' in i.get('path').upper(), file_indicators)
        )
        sha2_indicators = list(filter(lambda i: 'SHA-1' in i.get('path').upper(), file_indicators))
        md5_indicators = list(filter(lambda i: 'MD5' in i.get('path').upper(), file_indicators))

        if len(file_indicators) > 0:
            if (
                len(file_indicators) <= 3
                and len(sha256_indicators) <= 1
                and len(sha2_indicators) <= 1
                and len(md5_indicators) <= 1
            ):
                value = ' : '.join([v.get('value') for v in file_indicators])
                yield {
                    'type': 'File',
                    'summary': value,
                    'xid': Batch.generate_xid([signature.get('xid'), value]),
                    'associatedGroups': [{'groupXid': signature.get('xid')}],
                }
            else:
                for i in file_indicators:
                    yield {
                        'type': 'File',
                        'summary': i.get('value'),
                        'xid': Batch.generate_xid([signature.get('xid'), i.get('value')]),
                        'associatedGroups': [{'groupXid': signature.get('xid')}],
                    }


class STIXListener(STIXPatternListener):
    """Visitor for the parsed stix pattern."""

    def __init__(self):
        """Visitor for the parsed stix pattern."""
        super().__init__()
        self._indicators = []  # indicators that have been pulled out of this

    def enterPropTestEqual(self, ctx: STIXPatternParser.PropTestEqualContext):
        """Pull out path and value from statements with =.

        Args:
            ctx: the context of the equals statement.
        """
        test = ctx.getText()
        eq_index = test.index('=')
        if eq_index:
            path, value = test[:eq_index], test[(eq_index + 1) :]  # noqa: E203

            self._indicators.append({'path': path.strip(), 'value': value.strip()[1:-1]})

    def enterPropTestSet(self, ctx: STIXPatternParser.PropTestParenContext):
        """Pull out path and value from statements with in (...).

        Args:
            ctx: the context of the equals statement.
        """
        text = ctx.getText()
        path, values = text.split('IN')

        path = path.strip()

        values = values[1:-1]  # strip off the surrounding parens.
        values = [v.strip()[1:-1] for v in values.split(',')]  # split on , and strip

        for value in values:
            self._indicators.append({'path': path, 'value': value})

    @property
    def indicators(self):
        """Return the indicators parsed out of this pattern."""
        return self._indicators



"""Validators module for TcEx Framework"""
# flake8: noqa
from .in_range import in_range
from .not_in import not_in
from .operators import equal_to, greater_than, greater_than_or_equal, less_than, less_than_or_equal
from .transforms import to_bool, to_float, to_int
from .validation_exception import ValidationError



"""Validate Data Testing Module"""
# standard library
import datetime
import difflib
import gzip
import hashlib
import json
import math
import numbers
import operator
import os
import random
import re
from collections import OrderedDict
from typing import Optional, Union
from urllib.parse import quote, unquote

from ..utils import Utils


class Validator:
    """Validator"""

    def __init__(self, tcex: object, log: object):
        """Initialize class properties.

        Args:
            tcex: A configure instance of tcex.
            log: A configure instance of logger.
        """
        self.log = log
        self.tcex = tcex
        # TODO: validate this
        self.tcex.logger.update_handler_level('error')

        # properties
        self._redis = None
        self._threatconnect = None
        self.max_diff = 10
        self.truncate = 500
        self.utils = Utils()

    @staticmethod
    def _string_to_int_float(x: str) -> Union[float, int]:
        """Take string input and return float or int.

        Args:
            x: The value to coerce to float/int.

        Returns:
            Union[float, int]: The coerced valued.
        """
        if isinstance(x, bytes):
            x = x.decode('utf-8')

        try:
            f = float(x)
            i = int(f)
        except TypeError:
            return x  # return original value
        except ValueError:
            return x  # return original value
        else:
            if f != i:
                return f  # return float
            return i  # return int

    def compare(
        self,
        app_data: Union[dict, list, str],
        test_data: Union[dict, list, str],
        op: Optional[str] = None,
        **kwargs,
    ) -> tuple:
        """Compare app_data to test data.

        Args:
            app_data (dict, list, str): The data from the App.
            test_data (dict, list, str): The validation data from the test case.
            op (str, optional): The comparison operator expression. Defaults to "eq".

        Returns:
            [type]: [description]
        """
        # remove comment field from kwargs if it exists
        try:
            del kwargs['comment']
        except KeyError:
            pass

        op = op or 'eq'
        if not self.get_operator(op):
            self.log.data(
                'validate', 'Invalid Operator', f'Provided operator of {op} is invalid', 'error'
            )
            return False

        # logging header
        title = kwargs.pop('title', app_data)
        self.log.title(title, '=')

        # allow dev to provide formatted data for logging
        log_app_data = kwargs.pop('log_app_data', app_data)
        log_test_data = kwargs.pop('log_test_data', test_data)

        # run operator
        passed, details = self.get_operator(op)(app_data, test_data, **kwargs)

        # log validation data in a readable format
        self.validate_log_output(passed, log_app_data, log_test_data, details.strip(), op)

        # build assert error
        assert_error = (
            f'\n App Data     : {app_data}\n'
            f' Operator     : {op}\n'
            f' Expected Data: {test_data}\n'
            f' Details      : {details}\n'
        )
        return passed, assert_error

    def details(self, app_data, test_data, op):
        """Return details about the validation."""
        details = ''
        if app_data is not None and test_data is not None and op in ['eq', 'ne']:
            try:
                diff_count = 0
                for i, diff in enumerate(difflib.ndiff(app_data, test_data)):
                    if diff[0] == ' ':  # no difference
                        continue

                    if diff[0] == '-':
                        details += f'\n    * Missing data at index {i}'
                    elif diff[0] == '+':
                        details += f'\n    * Extra data at index {i}'
                    if diff_count > self.max_diff:
                        details += '\n    * Max number of differences reached.'
                        # don't spam the logs if string are vastly different
                        self.log.data(
                            'validate', 'Maximum Reached', 'Max number of differences reached.'
                        )
                        break
                    diff_count += 1
            except TypeError:
                pass
            except KeyError:
                pass
        return details

    def get_operator(self, op):
        """Get the corresponding operator"""
        operators = {
            'date_format': self.operator_date_format,
            'df': self.operator_date_format,
            'dd': self.operator_deep_diff,
            'is_url': self.operator_is_url,
            'is_date': self.operator_is_date,
            'is_number': self.operator_is_number,
            'length_eq': self.operator_length_eq,
            'leq': self.operator_length_eq,
            'eq': self.operator_eq,
            '=': self.operator_eq,
            'le': self.operator_le,
            '<=': self.operator_le,
            'lt': self.operator_lt,
            '<': self.operator_lt,
            'ge': self.operator_ge,
            '>=': self.operator_ge,
            'gt': self.operator_gt,
            '>': self.operator_gt,
            'jeq': self.operator_json_eq,
            'json_eq': self.operator_json_eq,
            'kveq': self.operator_keyvalue_eq,
            'keyvalue_eq': self.operator_keyvalue_eq,
            'ne': self.operator_ne,
            '!=': self.operator_ne,
            'rex': self.operator_regex_match,
            'skip': self.operator_skip,
        }
        return operators.get(op, None)

    def operator_date_format(self, app_data, test_data):
        """Check if date or dates match the provide format.

        Args:
            app_data: (list,str): One or more date strings.
            test_data: (str): A strptime string for comparision.
        """
        if self.check_null(test_data):
            return False, f'Invalid test_data: {test_data}. One or more values in test_data is null'
        if self.check_null(app_data):
            return False, f'Invalid app_data: {app_data}. One or more values in app_data is null'

        if not isinstance(app_data, list):
            app_data = [app_data]
        bad_data = []
        passed = True
        for data in app_data:
            try:
                datetime.datetime.strptime(data, test_data)
            except ValueError:
                bad_data.append(data)
                passed = False
        return passed, ','.join(bad_data)

    @staticmethod
    def operator_deep_diff(
        app_data, test_data, **kwargs
    ):  # pylint: disable=too-many-return-statements
        """Compare app data equals tests data.

        Args:
            app_data (dict|str|list): The data created by the App.
            test_data (dict|str|list): The data provided in the test case.

        Returns:
            bool: The results of the operator.
        """
        try:
            # third-party
            from deepdiff import DeepDiff
        except ImportError:
            return False, 'Could not import DeepDiff module (try "pip install deepdiff").'

        if (test_data is None or app_data is None) and app_data != test_data:
            # handle "null" -> None match
            return False, f'App Data {app_data} does not match Test Data {test_data}'

        # deepdiff doesn't handle ordered dicts properly
        safe_app_data = app_data
        if isinstance(app_data, OrderedDict):
            safe_app_data = json.loads(json.dumps(app_data))
        elif isinstance(app_data, list):
            safe_app_data = []
            for ad in app_data:
                if isinstance(ad, OrderedDict):
                    ad = json.loads(json.dumps(ad))
                safe_app_data.append(ad)

        # deepdiff doesn't handle ordered dicts properly
        safe_test_data = test_data
        if isinstance(test_data, OrderedDict):
            safe_test_data = json.loads(json.dumps(test_data))
        elif isinstance(test_data, list):
            safe_test_data = []
            for td in test_data:
                if isinstance(td, OrderedDict):
                    td = json.loads(json.dumps(td))
                safe_test_data.append(td)

        # run operator
        try:
            ddiff = DeepDiff(safe_app_data, safe_test_data, **kwargs)
        except KeyError:
            return False, 'Encountered KeyError when running deepdiff'
        except NameError:
            return False, 'Encountered NameError when running deepdiff'

        if ddiff:
            return False, str(ddiff)
        return True, ''

    def operator_eq(self, app_data, test_data):
        """Compare app data is equal to tests data.

        Args:
            app_data (dict, list, str): The data created by the App.
            test_data (dict, list, str): The data provided in the test case.

        Returns:
            bool, str: The results of the operator and any error message
        """
        results = operator.eq(app_data, test_data)
        return results, self.details(app_data, test_data, 'eq')

    def operator_is_url(self, app_data, test_data):  # pylint: disable=unused-argument
        """Check if the app_data is a known date."""
        if self.check_null(app_data):
            return False, f'Invalid app_data: {app_data}. One or more values in app_data is null'

        if not isinstance(app_data, list):
            app_data = [app_data]
        bad_data = []
        passed = True
        regex = re.compile(
            r'^(?:http|ftp)s?://'  # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|[/?]\S+)$',
            re.IGNORECASE,
        )
        for data in app_data:
            try:
                matched = re.match(regex, data)
                if not matched:
                    bad_data.append(data)
                    passed = False
            except RuntimeError:
                bad_data.append(data)
                passed = False
        return passed, ','.join(bad_data)

    def operator_ge(self, app_data, test_data):
        """Compare app data is greater than or equal to tests data.

        Args:
            app_data (dict, list, str): The data created by the App.
            test_data (dict, list, str): The data provided in the test case.

        Returns:
            bool, str: The results of the operator and any error message
        """
        if self.check_null(test_data):
            return False, f'Invalid test_data: {test_data}. One or more values in test_data is null'
        if self.check_null(app_data):
            return False, f'Invalid app_data: {app_data}. One or more values in app_data is null'

        app_data = self._string_to_int_float(app_data)
        test_data = self._string_to_int_float(test_data)
        results = operator.ge(app_data, test_data)
        details = ''
        if not results:
            details = f'{app_data} {type(app_data)} !(>=) {test_data} {type(test_data)}'
        return results, details

    def operator_gt(self, app_data, test_data):
        """Compare app data is greater than tests data.

        Args:
            app_data (dict, list, str): The data created by the App.
            test_data (dict, list, str): The data provided in the test case.

        Returns:
            bool, str: The results of the operator and any error message
        """
        if self.check_null(test_data):
            return False, f'Invalid test_data: {test_data}. One or more values in test_data is null'
        if self.check_null(app_data):
            return False, f'Invalid app_data: {app_data}. One or more values in app_data is null'

        app_data = self._string_to_int_float(app_data)
        test_data = self._string_to_int_float(test_data)
        results = operator.gt(app_data, test_data)
        details = ''
        if not results:
            details = f'{app_data} {type(app_data)} !(>) {test_data} {type(test_data)}'
        return results, details

    def operator_is_date(self, app_data, test_data):  # pylint: disable=unused-argument
        """Check if the app_data is a known date."""
        if self.check_null(test_data):
            return False, f'Invalid test_data: {test_data}. One or more values in test_data is null'
        if self.check_null(app_data):
            return False, f'Invalid app_data: {app_data}. One or more values in app_data is null'

        if not isinstance(app_data, list):
            app_data = [app_data]
        bad_data = []
        passed = True
        for data in app_data:
            try:
                self.tcex.utils.datetime.any_to_datetime(data)
            except RuntimeError:
                bad_data.append(data)
                passed = False
        return passed, ','.join(bad_data)

    @staticmethod
    def check_null(data_list):
        """Check if data_list is None or if None exists in the data_list."""
        if not isinstance(data_list, list):
            data_list = [data_list]
        for data in data_list:
            if data is None:
                return True
        return False

    def operator_is_number(self, app_data, test_data):  # pylint: disable=unused-argument
        """Check if the app_data is a number."""
        if self.check_null(app_data):
            return False, f'Invalid app_data: {app_data}. One or more values in app_data is null'

        if not isinstance(app_data, list):
            app_data = [app_data]
        bad_data = []
        passed = True
        for data in app_data:
            if isinstance(data, str) and isinstance(self._string_to_int_float(data), (int, float)):
                continue
            if isinstance(data, numbers.Number):
                continue
            bad_data.append(data)
            passed = False
        return passed, ','.join(bad_data)

    def operator_json_eq(self, app_data, test_data, **kwargs):
        """Compare app data equals tests data.

        Takes a str, dict, or list value and removed field before passing to deepdiff. Only fields
        at the "root" level can be removed (e.g., "date", not "data.date").

        Args:
            app_data (dict|str|list): The data created by the App.
            test_data (dict|str|list): The data provided in the test case.

        Returns:
            bool: The results of the operator.
        """
        if (test_data is None or app_data is None) and app_data != test_data:
            # handle "null" -> None match
            return False, f'App Data {app_data} does not match Test Data {test_data}'

        if isinstance(app_data, (str)):
            try:
                app_data = json.loads(app_data)
            except ValueError:
                return False, f'Invalid JSON data provide ({app_data}).'
        elif isinstance(app_data, (list)):
            # ADI-1076/ADI-1149
            app_data_updated = []
            for ad in app_data:
                try:
                    if isinstance(ad, (OrderedDict, dict)):
                        ad = json.dumps(ad)
                except ValueError:
                    return False, f'Invalid JSON data provide ({app_data}).'

                try:
                    # APP-599 - best effort try to stringify value in list
                    ad = json.loads(ad)
                except Exception:
                    pass
                app_data_updated.append(ad)
            app_data = app_data_updated

        if isinstance(test_data, (str)):
            try:
                test_data = json.loads(test_data)
            except ValueError:
                return False, f'Invalid JSON data provide ({test_data}).'
        elif isinstance(test_data, (list)):
            # ADI-1076/ADI-1149
            test_data_updated = []
            for td in test_data:
                try:
                    if isinstance(td, (OrderedDict, dict)):
                        td = json.dumps(td)
                except ValueError:
                    return False, f'Invalid JSON data provide ({test_data}).'

                try:
                    # APP-599 - best effort try to stringify value in list
                    td = json.loads(td)
                except Exception:
                    pass
                test_data_updated.append(td)
            test_data = test_data_updated

        exclude = kwargs.pop('exclude', [])
        if isinstance(app_data, list) and isinstance(test_data, list):
            app_data = [self.operator_json_eq_exclude(ad, exclude) for ad in app_data]
            test_data = [self.operator_json_eq_exclude(td, exclude) for td in test_data]
        elif isinstance(app_data, dict) and isinstance(test_data, dict):
            app_data = self.operator_json_eq_exclude(app_data, exclude)
            test_data = self.operator_json_eq_exclude(test_data, exclude)

        return self.operator_deep_diff(app_data, test_data, **kwargs)

    def operator_json_eq_exclude(self, data, exclude):
        """Remove excluded field from dictionary.

        Args:
            app_data (dict|str): The data to be processed.
            exclude (list): The key names to be "excluded" from data.

        Returns:
            dict: The data with excluded values removed.
        """
        for e in exclude:
            try:
                es = e.split('.')
                data = self.remove_excludes(data, es)
            except (AttributeError, KeyError, TypeError) as err:
                self.log.data(
                    'validate',
                    'Invalid Config',
                    f'Invalid validation configuration: ({err})',
                    'error',
                )
        return data

    def operator_keyvalue_eq(self, app_data, test_data, **kwargs):
        """Compare app data equals tests data.

        Args:
            app_data (dict|str|list): The data created by the App.
            test_data (dict|str|list): The data provided in the test case.

        Returns:
            bool: The results of the operator.
        """
        # remove exclude_key field. usually dynamic data like date fields.
        if kwargs.get('exclude_keys') is not None:
            app_data = [
                kv for kv in app_data if kv.get('key') not in kwargs.get('exclude_keys', [])
            ]
            test_data = [
                kv for kv in test_data if kv.get('key') not in kwargs.get('exclude_keys', [])
            ]
            del kwargs['exclude_keys']

        return self.operator_deep_diff(app_data, test_data, **kwargs)

    def operator_le(self, app_data, test_data):
        """Compare app data is less than or equal to tests data.

        Args:
            app_data (dict, list, str): The data created by the App.
            test_data (dict, list, str): The data provided in the test case.

        Returns:
            bool, str: The results of the operator and any error message
        """
        if self.check_null(test_data):
            return False, f'Invalid test_data: {test_data}. One or more values in test_data is null'
        if self.check_null(app_data):
            return False, f'Invalid app_data: {app_data}. One or more values in app_data is null'

        app_data = self._string_to_int_float(app_data)
        test_data = self._string_to_int_float(test_data)
        results = operator.le(app_data, test_data)
        details = ''
        if not results:
            details = f'{app_data} {type(app_data)} !(<=) {test_data} {type(test_data)}'
        return results, details

    @staticmethod
    def operator_length_eq(app_data, test_data):
        """Check length of app_data.

        If data passed in is 2 lists, validates length lists are the same.
        If data passed in is 2 strings, validates length strings are the same.
        If data passed in is 1 list and 1 int, validates length array and int value are the same.
        If data passed in is 1 str and 1 int, validates length str and int value are the same.
        """
        if app_data is None:
            return False, f'Invalid test_data: {app_data}. Value in app_data is null'
        if test_data is None:
            return False, f'Invalid test_data: {test_data}. Value in test_data is null'

        if not (
            (isinstance(test_data, str) and isinstance(app_data, str))
            or (isinstance(test_data, list) and isinstance(app_data, list))
            or (isinstance(test_data, int) and isinstance(app_data, (list, str)))
        ):
            msg = (
                f'Cannot compare App Data Type: {type(app_data)} '
                f'to Test Data Type {type(test_data)}'
            )
            return False, msg
        app_len = len(app_data)
        if isinstance(test_data, int):
            test_len = test_data
        else:
            test_len = len(test_data)

        results = operator.eq(app_len, test_len)
        return results, f'App Data Length: {app_len} | Test Data Length: {test_len}'

    def operator_lt(self, app_data, test_data):
        """Compare app data is less than tests data.

        Args:
            app_data (dict, list, str): The data created by the App.
            test_data (dict, list, str): The data provided in the test case.

        Returns:
            bool, str: The results of the operator and any error message
        """
        if self.check_null(test_data):
            return False, f'Invalid test_data: {test_data}. One or more values in test_data is null'
        if self.check_null(app_data):
            return False, f'Invalid app_data: {app_data}. One or more values in app_data is null'

        app_data = self._string_to_int_float(app_data)
        test_data = self._string_to_int_float(test_data)
        results = operator.lt(app_data, test_data)
        details = ''
        if not results:
            details = f'{app_data} {type(app_data)} !(<) {test_data} {type(test_data)}'
        return results, details

    def operator_ne(self, app_data, test_data):
        """Compare app data is not equal to tests data.

        Args:
            app_data (dict, list, str): The data created by the App.
            test_data (dict, list, str): The data provided in the test case.

        Returns:
            bool, str: The results of the operator and any error message
        """
        results = operator.ne(app_data, test_data)
        return results, self.details(app_data, test_data, 'eq')

    def operator_regex_match(self, app_data, test_data):
        """Compare app data matches test regex data.

        Args:
            app_data (str, list): The data created by the App.
            test_data (str): The data provided in the test case.

        Returns:
            bool: The results of the operator.
        """
        if self.check_null(test_data):
            return False, f'Invalid test_data: {test_data}. One or more values in test_data is null'
        if self.check_null(app_data):
            return False, f'Invalid app_data: {app_data}. One or more values in app_data is null'

        if isinstance(app_data, dict):
            return False, 'Invalid app_data, dict type is not supported.'

        if not isinstance(app_data, list):
            app_data = [app_data]
        bad_data = []
        passed = True
        for data in app_data:
            if re.match(test_data, data) is None:
                bad_data.append(data)
                passed = False

        bad_data = ','.join(bad_data)  # convert bad_data to string for assertion error
        if bad_data:
            bad_data = f'Failed inputs: {bad_data}'
        return passed, bad_data

    @staticmethod
    def operator_skip(app_data, test_data):  # pylint: disable=unused-argument
        """Skip validation and always return True.

        Args:
            app_data (str, list): The data created by the App.
            test_data (str): The data provided in the test case.

        Returns:
            bool: The results of the operator.
        """
        return True, 'skipped'

    @property
    def redis(self):
        """Get the current instance of Redis for validating data"""
        if not self._redis:
            self._redis = Redis(self)
        return self._redis

    def remove_excludes(self, dict_1, paths):
        """Remove a list of paths from a given dict

        ex:
            dict_1: {
                'result': {
                    'sys_id': 123,
                    'owner': {
                        'id': 5,
                        'name': 'System'
                    },
                    'name': results
                },
                'status': 'Uploaded
            paths: ['result', 'owner', 'id']

            returns: {
                'result': {
                    sys_id': 123,
                        'owner': {
                            'name': 'System'
                        },
                    'name': results
                },
                'status': 'Uploaded
            }
        """
        if isinstance(dict_1, list):
            for item in dict_1:
                self.remove_excludes(item, paths)
        elif not isinstance(dict_1, dict):
            raise RuntimeError(f'Provided value ({dict_1}) must be a dict.')

        path_0 = paths[0]
        if len(paths) == 1:
            dict_1.pop(path_0, None)
            return dict_1
        self.remove_excludes(dict_1.get(path_0, {}), paths[1:])
        return dict_1

    @property
    def threatconnect(self):
        """Get the current instance of ThreatConnect for validating data"""
        if not self._threatconnect:
            self._threatconnect = ThreatConnect(self)
        return self._threatconnect

    def validate_log_output(self, passed, app_data, test_data, details, op):
        """Format the validation log output to be easier to read.

        Args:
            passed (bool): The results of the validation test.
            app_data (str): The data store in Redis.
            test_data (str): The user provided data.
            op (str): The comparison operator.

        Raises:
            RuntimeError: Raise error on validation failure if halt_on_fail is True.
        """
        truncate = self.truncate
        if app_data is not None and passed:
            if isinstance(app_data, (str)) and len(app_data) > truncate:
                app_data = app_data[:truncate]
            elif isinstance(app_data, (list)):
                db_data_truncated = []
                for d in app_data:
                    if d is not None and isinstance(d, str) and len(d) > truncate:
                        db_data_truncated.append(f'{d[:self.truncate]} ...')
                    else:
                        db_data_truncated.append(d)
                app_data = db_data_truncated

        if test_data is not None and passed:
            if isinstance(test_data, (str)) and len(test_data) > truncate:
                test_data = test_data[: self.truncate]
            elif isinstance(test_data, (list)):
                user_data_truncated = []
                for u in test_data:
                    if isinstance(app_data, (str)) and len(u) > truncate:
                        user_data_truncated.append(f'{u[:self.truncate]} ...')
                    else:
                        user_data_truncated.append(u)
                test_data = user_data_truncated

        self.log.data('validate', 'App Data', f'({app_data}), Type: [{type(app_data)}]')
        self.log.data('validate', 'Operator', op)
        self.log.data('validate', 'Test Data', f'({test_data}), Type: [{type(test_data)}]')

        if passed:
            self.log.data('validate', 'Result', 'Passed')
        else:
            self.log.data('validate', 'Result', 'Failed', 'error')
            self.log.data('validate', 'Details', details, 'error')


class Redis:
    """Validates Redis data"""

    def __init__(self, provider):
        """Initialize class properties."""
        self.provider = provider

        # Properties
        self.log = self.provider.log
        self.redis_client = provider.tcex.redis_client

    def not_null(self, variable):
        """Validate that a variable is not empty/null"""
        # Could do something like self.ne(variable, None), but want to be pretty specific on
        # the errors on this one
        variable_data = self.provider.tcex.playbook.read(variable)
        self.log.data('validate', 'Variable', variable)
        self.log.data('validate', 'DB Data', variable_data)
        if not variable:
            self.log.data(
                'validate', 'None Error', 'Redis variable not provided', 'error',
            )
            return False

        if not variable_data:
            self.log.data(
                'validate', 'None Found', f'Redis variable {variable} was not found', 'error',
            )
            return False

        return True

    def type(self, variable):
        """Validate the type of a redis variable"""
        variable_data = self.provider.tcex.playbook.read(variable)
        self.log.data('validate', 'Variable', variable)
        self.log.data('validate', 'DB Data', variable_data)
        redis_type = self.provider.tcex.playbook.variable_type(variable)
        if redis_type.endswith('Array'):
            redis_type = list
        elif redis_type.startswith('String'):
            redis_type = str
        elif redis_type.startswith('KeyValuePair'):
            redis_type = dict
        else:
            redis_type = str

        if not variable_data:
            self.log.data(
                'validate', 'None Found', f'Redis variable {variable} was not found', 'error',
            )
            return False
        if not isinstance(variable_data, redis_type):
            self.log.data(
                'validate',
                'Type Mismatch',
                f'Redis Type: {redis_type}  and Variable: {variable} do not match',
                'error',
            )
            return False

        return True

    def data(self, variable, test_data, op=None, **kwargs):
        """Validate Redis data <operator> test_data.

        Args:
            variable (str): The variable to read from REDIS.
            data (dict or list or str): The validation data
            op (str, optional): The comparison operator expression. Defaults to "eq".

        Returns:
            [type]: [description]
        """
        # remove comment field from kwargs if it exists
        try:
            del kwargs['comment']
        except KeyError:
            pass

        # log skipped validations here so that variable can be logged
        if op == 'skip':
            self.log.data('validate', 'skipped validation', variable, 'warning')

        op = op or 'eq'
        if not variable:
            self.log.data(
                'validate', 'None Error', 'Redis variable not provided', 'error',
            )
            return False

        if not self.provider.get_operator(op):
            self.log.error(f'Invalid operator provided ({op})')
            self.log.data(
                'validate', 'Invalid Operator', f'Provided operator of {op} is invalid', 'error'
            )
            return False

        if variable.endswith('Binary'):
            app_data = self.provider.tcex.playbook.read_binary(variable, False, False)
        elif variable.endswith('BinaryArray'):
            app_data = self.provider.tcex.playbook.read_binary_array(variable, False, False)
        else:
            app_data = self.provider.tcex.playbook.read(variable)

        # logging header
        self.log.title(variable, '-')

        # run operator
        passed, details = self.provider.get_operator(op)(app_data, test_data, **kwargs)

        # log validation data in a readable format
        self.provider.validate_log_output(passed, app_data, test_data, details.strip(), op)

        # TODO: move this to a value_rule_validator method so that
        # more check like this can be performed

        # check for bad string values
        suspect_values = ['False', 'null', 'None', 'True']
        if app_data in suspect_values:
            self.log.data(
                'validate',
                'Suspect Value',
                f'App data matched a suspect value ({suspect_values}).',
                'warning',
            )

        # ADI-1118 - log warning if string value is wrapped in quotes
        if isinstance(app_data, (str)) and app_data.startswith('"') and app_data.endswith('"'):
            self.log.data(
                'validate', 'Suspect Value', 'App data is wrapped in double quotes.', 'warning',
            )

        # build assert error
        assert_error = (
            f'\n Variable      : {variable}\n'
            f' App Data      : {app_data}\n'
            f' Operator      : {op}\n'
            f' Expected Data : {test_data}\n'
        )
        if details:
            assert_error += f' Details       : {details}\n'
        return passed, assert_error

    def eq(self, variable, data):
        """Validate test data equality"""
        return self.data(variable, data)

    def dd(self, variable, data, **kwargs):
        """Validate test data equality"""
        return self.data(variable, data, op='dd', **kwargs)

    def ge(self, variable, data):
        """Validate test data equality"""
        return self.data(variable, data, op='ge')

    def gt(self, variable, data):
        """Validate test data equality"""
        return self.data(variable, data, op='gt')

    def jeq(self, variable, data, **kwargs):
        """Validate JSON data equality"""
        return self.data(variable, data, op='jeq', **kwargs)

    def json_eq(self, variable, data, **kwargs):
        """Validate JSON data equality"""
        return self.data(variable, data, op='jeq', **kwargs)

    def kveq(self, variable, data, **kwargs):
        """Validate JSON data equality"""
        return self.data(variable, data, op='kveq', **kwargs)

    def keyvalue_eq(self, variable, data, **kwargs):
        """Validate KeyValue JSON data equality"""
        return self.data(variable, data, op='kveq', **kwargs)

    def lt(self, variable, data):
        """Validate test data less than"""
        return self.data(variable, data, op='lt')

    def le(self, variable, data):
        """Validate test data less than or equal"""
        return self.data(variable, data, op='le')

    def ne(self, variable, data):
        """Validate test data non equality"""
        return self.data(variable, data, op='ne')

    def rex(self, variable, data):
        """Test App data with regex"""
        return self.data(variable, fr'{data}', op='rex')

    def skip(self, variable, data):
        """Test App data with regex"""
        return self.data(variable, data, op='skip')


class ThreatConnect:
    """Validate ThreatConnect data"""

    def __init__(self, provider):
        """Initialize class properties"""
        self.provider = provider
        self.log = self.provider.log

    def batch(self, profile):
        """Validate the batch submission"""
        batch_base_path = os.path.join(
            self.provider.tcex.args.tc_temp_path,
            profile.feature,
            f'test_profiles-{profile.name}',
            'DEBUG',
        )
        batch_data_path = os.path.join(batch_base_path, 'batch_data',)
        batch_file_path = os.path.join(batch_base_path, 'batch_files',)

        validation_percent = profile.validation_criteria.get('percent')
        validation_count = profile.validation_criteria.get('count')
        if not validation_count and not validation_percent:
            self.log.info(f'action=validate-batch, percent={validation_percent}')
            self.log.data(
                'validate', 'Batch Submission', 'Skipping validation due to no criteria provided',
            )
            return

        batch_submit_totals = self._get_batch_submit_totals(batch_data_path)

        if validation_count:
            validation_percent = self._convert_to_percent(validation_count, batch_submit_totals)
        self.log.info(
            f'''action=validate-batch, count={validation_count}, percent={validation_percent}'''
        )

        batch_errors = []
        for filename in os.listdir(batch_data_path):
            if not filename.startswith('errors-'):
                continue

            with gzip.open(os.path.join(batch_data_path, filename)) as fh:
                batch_errors += json.load(fh)

        for filename in os.listdir(batch_data_path):
            if not filename.startswith('batch-'):
                continue

            with gzip.open(os.path.join(batch_data_path, filename)) as fh:
                data = json.load(fh)
                validation_data = self._partition_batch_data(data)
                sample_validation_data = []
                for key in validation_data:
                    for sub_partition in validation_data.get(key).values():
                        sample_size = math.ceil(len(sub_partition) * (validation_percent / 100))
                        sample_validation_data.extend(random.sample(sub_partition, sample_size))
                self.log.info(f'validating {len(sample_validation_data)} entries')

                files = []
                for sample_data in sample_validation_data:
                    sample_data_type = sample_data.get('type').lower()
                    if sample_data_type not in ['document', 'report']:
                        files.append(None)
                        continue

                    sample_data_type = 'reports'
                    if sample_data_type == 'document':
                        sample_data_type = 'documents'

                    filename = (
                        sample_data_type
                        + '--'
                        + sample_data.get('xid')
                        + '--'
                        + sample_data.get('name', '')
                    )
                    filename = os.path.join(batch_file_path, filename)
                    if os.path.isfile(filename):
                        files.append(filename)
                    else:
                        files.append(None)
                results = self.tc_entities(sample_validation_data, profile.owner, files=files)
                for result in results:
                    if result.get('valid'):
                        continue
                    name = result.get('name')
                    batch_error = self._batch_error(name, batch_errors)
                    if batch_error:
                        self.log.data(
                            'validate',
                            'Batch Submission',
                            f'Errors validating {name} ({batch_error})h',
                            'error',
                        )
                        continue
                    # TODO: Should use pip install pytest-check is_true method so it wont fail after
                    # one failed assert.
                    assert result.get('valid'), (
                        f'{name} in ThreatConnect did not match what was submitted. '
                        f"Errors:{result.get('errors')}"
                    )
        self._log_batch_submit_details(batch_errors, profile.owner)

    def _log_batch_submit_details(self, batch_errors, owner):
        if not batch_errors:
            return

        counts = {}
        error_regex = r'\((.*?)\)'
        for error in batch_errors:
            reason = error.get('errorReason', '')
            m = re.search(error_regex, reason)
            if not m:
                continue
            if not m.group(1) in counts:
                counts[m.group(1)] = 0
            counts[m.group(1)] += 1
        log_message = ''
        for code, count in counts.items():
            log_message += (
                self.provider.tcex.batch(owner).error_codes.get(code, 'General Error')
                + ': '
                + str(count)
                + '\n'
            )

        self.log.data('validate', 'Batch Submission', log_message, 'error')

    @staticmethod
    def _batch_error(key, batch_errors):
        for error in batch_errors:
            reason = error.get('errorReason', '')
            if key in reason:
                return reason
        return None

    def dir(self, directory, owner):
        """Validate the content of a given dir"""
        results = []
        for test_file in os.listdir(directory):
            if not (test_file.endswith('.json') and test_file.startswith('validate_')):
                continue
            results.append(self.file(f'{directory}/{test_file}', owner))
        return results

    def file(self, file, owner):
        """Validate the content of a given file"""
        entities = self._convert_to_entities(file)
        return self.tc_entities(entities, owner)

    def tc_entities(self, tc_entities, owner, files=None):
        """Validate a array of tc_entities"""
        results = []
        if files:
            if not len(tc_entities) == len(files):
                return [
                    {
                        'valid': False,
                        'errors': [
                            'LengthError: Length of files provided does not '
                            'match length of entities provided.'
                        ],
                    }
                ]

        for index, entity in enumerate(tc_entities):
            name = entity.get('name', entity.get('summary'))
            if files:
                valid, errors = self.tc_entity(entity, owner, files[index])
            else:
                valid, errors = self.tc_entity(entity, owner)
            results.append({'name': name, 'valid': valid, 'errors': errors})
        return results

    def tc_entity(self, tc_entity, owner, file=None):  # pylint: disable=unused-argument
        """Validate the ti_response entity"""
        parameters = {'includes': ['additional', 'attributes', 'labels', 'tags']}
        ti_entity = self._convert_to_ti_entity(tc_entity, owner)
        if ti_entity is None:
            error = (
                f"NotFoundError: Provided {tc_entity.get('type')}: "
                f"{tc_entity.get('summary', tc_entity.get('name', 'Unknown'))} could not "
                f'be fetched from ThreatConnect owner {owner} (Null Means default API Owner).'
            )
            return False, [error]
        ti_response = ti_entity.single(params=parameters)
        entity_name = tc_entity.get('name')
        errors = []
        if not self.success(ti_response):
            error = (
                f"NotFoundError: Provided {tc_entity.get('type')}: "
                f"{tc_entity.get('summary', entity_name)} could not "
                f'be fetched from ThreatConnect'
            )
            return False, [error]

        ti_response_entity = None
        ti_response = ti_response.json().get('data', {}).get(ti_entity.api_entity, {})
        for entity in self.provider.tcex.ti.entities(ti_response, tc_entity.get('type', None)):
            ti_response_entity = entity
            # pylint: disable=unused-variable
            valid_attributes, attributes_errors = self._response_attributes(ti_response, tc_entity)
            # pylint: disable=unused-variable
            valid_tags, tag_errors = self._response_tags(ti_response, tc_entity)
            # pylint: disable=unused-variable
            valid_labels, label_errors = self._response_labels(ti_response, tc_entity)
            # pylint: disable=unused-variable
            valid_file, file_errors = self._file(ti_entity, file)

            errors = attributes_errors + tag_errors + label_errors + file_errors

        if ti_entity.type == 'Indicator':
            provided_rating = tc_entity.get('rating', None)
            expected_rating = ti_response.get('rating', None)
            if not provided_rating == expected_rating:
                errors += (
                    f'RatingError: Provided rating {provided_rating} '
                    f'does not match actual rating {expected_rating}'
                )

            provided_confidence = tc_entity.get('confidence', None)
            expected_confidence = ti_response.get('confidence', None)
            if not provided_confidence == expected_confidence:
                errors += (
                    f'ConfidenceError: Provided confidence {provided_confidence} '
                    f'does not match actual confidence {expected_confidence}'
                )

            provided_summary = unquote(
                ':'.join([x for x in ti_entity.unique_id.split(':') if x.strip()])
            )
            expected_summary = unquote(ti_response_entity.get('value', ''))
            if provided_summary != expected_summary:
                errors += (
                    f'SummaryError: Provided summary {provided_summary} '
                    f'does not match actual summary {expected_summary}'
                )
        elif ti_entity.type == 'Group':
            provided_summary = tc_entity.get('name', None)
            expected_summary = ti_response_entity.get('value', None)
            if not provided_summary == expected_summary:
                errors += (
                    f'SummaryError: Provided summary {provided_summary} '
                    f'does not match actual summary {expected_summary}'
                )

        return not bool(errors), errors

    def flatten(self, lis):
        """Idk why python doesn't have this built in but helper function to flatten a list"""
        new_lis = []
        for item in lis:
            if isinstance(item, list):
                new_lis.extend(self.flatten(item))
            else:
                new_lis.append(item)
        return new_lis

    @staticmethod
    def compare_dicts(expected, actual, error_type=''):
        """Compare two dicts and returns a list of errors if they don't match"""
        errors = []

        for item in expected:
            if item in actual:
                if isinstance(expected.get(item), list):
                    actual_list_string = [str(i) for i in actual.get(item, [])]
                    for value in expected.get(item):
                        if str(value) not in actual_list_string:
                            errors.append(
                                f'{error_type}{item} : {expected.get(item)} '
                                f'did not match {item} : {actual.get(item)}'
                            )
                else:
                    if str(expected.get(item)) != actual.get(item):
                        errors.append(
                            f'{error_type}{item} : {expected.get(item)} '
                            f'did not match {item} : {actual.get(item)}'
                        )
                actual.pop(item)
            else:
                errors.append(
                    f'{error_type}{item} : {expected.get(item)} was '
                    f'in expected results but not in actual results.'
                )
        for item in list(actual.items()):
            errors.append(
                f'{error_type}{item} : {item} was in '
                f'actual results but not in expected results.'
            )

        return bool(errors), errors

    @staticmethod
    def compare_lists(expected, actual, error_type=''):
        """Compare two lists and returns a list of errors if they don't match"""
        errors = []
        for item in expected:
            if item in actual:
                actual.remove(item)
            else:
                errors.append(
                    f'{error_type}{item} was in expected results but not in actual results.'
                )
        for item in actual:
            errors.append(f'{error_type}{item} was in actual results but not in expected results.')

        return bool(errors), errors

    @staticmethod
    def _convert_to_entities(file):
        """Convert file to tc_entity array"""
        with open(file) as read_file:
            data = json.load(read_file)
        return data

    def _custom_indicators(self):
        indicator_details = self.provider.tcex.indicator_types_data
        custom_indicators = []
        for indicator in indicator_details.values():
            if indicator.get('custom', False):
                custom_indicators.append(indicator.get('name'))

        return custom_indicators

    def _convert_to_ti_entity(self, tc_entity, owner):
        """Convert a tc_entity to a ti_entity"""
        ti_entity = None

        if tc_entity.get('type').lower() in map(str.lower, self.provider.tcex.group_types):
            # We can't search by xid sadly so have to search by name and validate xid to
            # get the id of the group.
            filters = self.provider.tcex.ti.filters()
            filters.add_filter('name', '=', tc_entity.get('name'))
            generic_group_entity = self.provider.tcex.ti.group(
                group_type=tc_entity.get('type'), owner=owner
            )
            parameters = {'includes': ['additional']}
            for entity in generic_group_entity.many(filters=filters, params=parameters):
                if entity.get('xid') == tc_entity.get('xid'):
                    ti_entity = self.provider.tcex.ti.group(
                        group_type=tc_entity.get('type'),
                        owner=owner,
                        name=entity.get('name'),
                        unique_id=entity.get('id'),
                    )
        elif tc_entity.get('type').lower() in map(str.lower, self.provider.tcex.indicator_types):
            tc_entity['summary'] = quote(tc_entity.get('summary'), safe='')
            if tc_entity.get('type').lower() == 'file':
                tc_entity['summary'] = tc_entity.get('summary').upper()
            ti_entity = self.provider.tcex.ti.indicator(
                indicator_type=tc_entity.get('type'),
                owner=owner,
                unique_id=tc_entity.get('summary'),
            )
        elif tc_entity.get('type').lower() == 'victim':
            # TODO: Will need to do something similar to what was done to get the groups entity.
            pass

        return ti_entity

    def _get_batch_submit_totals(self, batch_data_path):
        """Break the batch submitions up into separate partitions.

        Each partition containing its total.
        """
        counts = {}
        for filename in os.listdir(batch_data_path):
            if not filename.startswith('batch-'):
                continue
            with gzip.open(os.path.join(batch_data_path, filename)) as fh:
                data = json.load(fh)
                partitioned_data = self._partition_batch_data(data)
                for key in partitioned_data:
                    if key not in counts:
                        counts[key] = {}
                    for sub_key in partitioned_data.get(key):
                        if sub_key not in counts:
                            counts[key][sub_key] = 0
                        counts[key][sub_key] += len(partitioned_data[key][sub_key])
        return counts

    @staticmethod
    def _convert_to_percent(validation_count, batch_submit_totals):
        """Given a count, calculate what percentage of the batch submits that is"""
        total = 0
        for key in batch_submit_totals.keys():
            for count in batch_submit_totals.get(key).values():
                total += count
        if validation_count > total:
            return 100
        return round((validation_count / total) * 100, 2)

    @staticmethod
    def _partition_batch_data(data):
        """Partitions the batch submittions into batch_submit[type][subtype] partitions"""
        partitioned_data = {}
        for key in data.keys():
            partitioned_sub_data = {}
            for sub_data in data.get(key):
                if not sub_data.get('type') in partitioned_sub_data.keys():
                    partitioned_sub_data[sub_data.get('type')] = [sub_data]
                else:
                    partitioned_sub_data[sub_data.get('type')].append(sub_data)
            partitioned_data[key] = partitioned_sub_data
        return partitioned_data

    def _response_attributes(self, ti_response, tc_entity):
        """Validate the ti_response attributes"""
        if not ti_response or not tc_entity:
            return True

        expected = {}
        actual = {}
        for attribute in tc_entity.get('attribute', []):
            attribute_type = attribute.get('type')
            expected.setdefault(attribute_type, []).append(attribute.get('value'))
        for attribute in ti_response.get('attribute', []):
            attribute_type = attribute.get('type')
            actual.setdefault(attribute_type, []).append(attribute.get('value'))
        return self.compare_dicts(expected, actual, error_type='AttributeError: ')

    def _response_tags(self, ti_response, tc_entity):
        """Validate the ti_response tags"""
        if not ti_response or not tc_entity:
            return True

        expected = []
        actual = []
        for tag in tc_entity.get('tag', []):
            expected.append(tag.get('name').lower())
        for tag in ti_response.get('tag', []):
            actual.append(tag.get('name').lower())

        return self.compare_lists(expected, actual, error_type='TagError: ')

    def _response_labels(self, ti_response, tc_entity):
        """Validate the ti_response labels"""
        if not ti_response or not tc_entity:
            return True

        expected = []
        actual = []
        for tag in tc_entity.get('securityLabel', []):
            expected.append(tag.get('name'))
        for tag in ti_response.get('securityLabel', []):
            actual.append(tag.get('name'))

        return self.compare_lists(expected, actual, error_type='SecurityLabelError: ')

    @staticmethod
    def _file(ti_entity, file_):
        """Handle file data"""
        if not file_:
            return True, []

        errors = []
        if ti_entity.api_sub_type == 'Document' or ti_entity.api_sub_type == 'Report':
            actual_hash = ti_entity.get_file_hash()
            actual_hash = actual_hash.hexdigest()
            provided_hash = hashlib.sha256()
            with open(file_, 'rb') as f:
                for byte_block in iter(lambda: f.read(4096), b''):
                    provided_hash.update(byte_block)
            provided_hash = provided_hash.hexdigest()
            if not provided_hash == actual_hash:
                errors.append(
                    f'sha256 {provided_hash} of provided file did not '
                    f'match sha256 of actual file {actual_hash}'
                )
        return bool(errors), errors

    @staticmethod
    def success(r):
        """[summary]

        Args:
            r ([type]): [description]

        Returns:
            [type]: [description]
        """
        status = True
        if r.ok:
            try:
                if r.json().get('status') != 'Success':
                    status = False
            except Exception:
                status = False
        else:
            status = False
        return status



"""ThreatConnect SecurityLabel Object"""
# standard library
import json
from typing import Optional


class SecurityLabel:
    """ThreatConnect Batch SecurityLabel Object."""

    __slots__ = ['_label_data']

    def __init__(self, name: str, description: Optional[str] = None, color: Optional[str] = None):
        """Initialize Class Properties.

        Args:
            name: The value for this security label.
            description: A description for this security label.
            color: A color (hex value) for this security label.
        """
        self._label_data = {'name': name}
        # add description if provided
        if description is not None:
            self._label_data['description'] = description
        if color is not None:
            self._label_data['color'] = color

    @property
    def color(self) -> str:
        """Return Security Label color."""
        return self._label_data.get('color')

    @color.setter
    def color(self, color: str):
        """Set Security Label color."""
        self._label_data['color'] = color

    @property
    def data(self) -> dict:
        """Return Security Label data."""
        return self._label_data

    @property
    def description(self) -> str:
        """Return Security Label description."""
        return self._label_data.get('description')

    @description.setter
    def description(self, description: str):
        """Set Security Label description."""
        self._label_data['description'] = description

    @property
    def name(self) -> str:
        """Return Security Label name."""
        return self._label_data.get('name')

    def __str__(self) -> str:
        """Return string represtentation of object."""
        return json.dumps(self.data, indent=4)



"""TcEx Error Codes"""


class TcExErrorCodes:
    """TcEx Framework Error Codes."""

    @property
    def errors(self):
        """Return TcEx defined error codes and messages.

        .. note:: RuntimeErrors with a code of >= 10000 are considered critical.  Those < 10000
            are considered warning or errors and are up to the developer to determine the
            appropriate behavior.
        """
        return {
            # tcex general errors
            100: 'Generic error. See log for more details ({}).',
            105: 'Required Module is not installed ({}).',
            200: 'Failed retrieving Custom Indicator Associations types from API ({}).',
            210: 'Failure during token renewal ({}).',
            215: 'HMAC authorization requires a PreparedRequest Object.',
            220: 'Failed retrieving indicator types from API ({}).',
            # tcex resource
            300: 'Failed retrieving Bulk JSON ({}).',
            305: 'An invalid action/association name ({}) was provided.',
            350: 'Data Store request failed. API status code: {}, API message: {}.',
            # batch v2: 500-600
            520: 'File Occurrences can only be added to a File. Current type: {}.',
            540: 'Failed polling batch status ({}).',
            545: 'Failed polling batch status. API status code: {}, API message: {}.',
            550: 'Batch status check reached timeout ({} seconds).',
            560: 'Failed retrieving batch errors ({}).',
            580: 'Failed posting file data ({}).',
            585: 'Failed posting file data. API status code: {}, API message: {}.',
            590: 'No hash values provided.',
            # threat intelligence
            600: 'Failed adding group type "{}" with name "{}" ({}).',
            605: 'Failed adding attribute type "{}" with value "{}" to group id "{}" ({}).',
            610: 'Failed adding label "{}" to group id "{}" ({}).',
            615: 'Failed adding tag "{}" to group id "{}" ({}).',
            650: 'Failed adding label "{}" to attribute id "{}" ({}).',
            # metrics
            700: 'Failed to create metric. API status code: {}, API message: {}.',
            705: 'Error while finding metric by name. API status code: {}, API message: {}.',
            710: 'Failed to add metric data. API status code: {}, API message: {}.',
            715: 'No metric ID found for "{}".',
            # notifications
            750: 'Failed to send notification. API status code: {}, API message: {}.',
            # datastore
            800: 'Failed to create index. API status code: {}, API message: {}.',
            805: 'Failed to {} record data. API status code: {}, API message: {}.',
            # threat intelligence module
            905: 'Error during update. {} does not have a unique_id set and cannot be updated.',
            910: 'Error during get. {} does not have a unique_id set and cannot be fetched.',
            915: 'Error during delete. {} does not have a unique_id set and cannot be deleted.',
            920: (
                'Error during create. {} does not have required values set and cannot be '
                'created.'
            ),
            # TODO: fix this and all references
            925: 'Error invalid {}. {} does not accept that {}, {}: {}.',
            950: 'Error during pagination. API status code: {}, API message: {}, API Url: {}.',
            951: 'Error during {}. API status code: {}, API message: {}, API Url: {}.',
            952: 'Error during {}. API status code: {}, API message: {}, API Url: {}.',
            # batch v2 critical:
            10500: 'Critical batch error ({}).',
            10505: 'Failed submitting batch job requests ({}).',
            10510: 'Failed submitting batch job requests. API status code: {}, API message: {}.',
            10520: 'Failed submitting batch data ({}).',
            10525: 'Failed submitting batch data. API status code: {}, API message: {}.',
        }

    def message(self, code):
        """Return the error message.

        Args:
            code (integer): The error code integer.

        Returns:
            (string): The error message.
        """
        return self.errors.get(code)



"""Case Management API Endpoints"""
# standard library
from enum import Enum, unique


@unique
class ApiEndpoints(Enum):
    """Available API Endpoints for Case Management"""

    ARTIFACT_TYPES = '/v3/artifactTypes'
    ARTIFACTS = '/v3/artifacts'
    CASES = '/v3/cases'
    NOTES = '/v3/notes'
    TAGS = '/v3/tags'
    TASKS = '/v3/tasks'
    WORKFLOW_EVENTS = '/v3/workflowEvents'
    WORKFLOW_TEMPLATES = '/v3/workflowTemplates'



"""TcEx testing profile Class."""
# standard library
import json
import logging
import os
import re
import sys
from collections import OrderedDict
from random import randint

# third-party
import colorama as c

from ..app_config_object import InstallJson, LayoutJson, Permutations
from ..env_store import EnvStore
from ..sessions import TcSession
from ..utils import Utils
from .migrate import Migrate
from .populate import Populate
from .rules import Rules
from .session_manager import SessionManager

# autoreset colorama
c.init(autoreset=True, strip=False)


class Profile:
    """Testing Profile Class.

    Args:
        default_args (dict, optional): The default Args for the profile.
        feature (str, optional): The feature name.
        name (str, optional): The filename of the profile in the profile.d director.
        redis_client (redis.client.Redis, optional): An instance of Redis client.
        pytestconfig (?, optional): Pytest config object.
        monkeypatch (?, optional): Pytest monkeypatch object.
        tcex_testing_context (str, optional): The current context for this profile.
        logger (logging.Logger, optional): An logging instance.
        options (dict, optional): ?
    """

    def __init__(
        self,
        default_args=None,
        feature=None,
        name=None,
        redis_client=None,
        pytestconfig=None,
        monkeypatch=None,
        tcex_testing_context=None,
        logger=None,
        options=None,
    ):
        """Initialize Class properties."""
        self._default_args = default_args or {}
        self._feature = feature
        self._name = name
        self.log = logger or logging.getLogger('profile')
        self.redis_client = redis_client
        self.pytestconfig = pytestconfig
        self.monkeypatch = monkeypatch
        self.tcex_testing_context = tcex_testing_context
        self.test_options = options

        # properties
        self._app_path = os.getcwd()
        self._context_tracker = []
        self._data = None
        self._output_variables = None
        self._pytest_args = None
        self._session_manager = None
        self._session = None
        self.env_store = EnvStore(logger=self.log)
        self.ij = InstallJson(logger=self.log)
        self.lj = LayoutJson(logger=self.log)
        self.permutations = Permutations(logger=self.log)
        self.populate = Populate(self)
        self.rules = Rules(self)
        self.tc_staged_data = {}
        self.utils = Utils()

    @staticmethod
    def _flatten_inputs(inputs):
        """Flatten the inputs dict."""
        inputs_flattened = dict(inputs.get('defaults', {}))
        inputs_flattened.update(inputs.get('optional', {}))
        inputs_flattened.update(inputs.get('required', {}))
        return inputs_flattened

    @staticmethod
    def _sorted(data):
        """Return a sorted dict as an OrderedDict."""
        return json.loads(json.dumps(data, sort_keys=True), object_pairs_hook=OrderedDict)

    @property
    def _test_case_data(self):
        """Return partially parsed test case data."""
        return os.getenv('PYTEST_CURRENT_TEST').split(' ')[0].split('::')

    @property
    def _test_case_name(self):
        """Return partially parsed test case data."""
        return self._test_case_data[-1].replace('/', '-').replace('[', '-').replace(']', '')

    def add(self, profile_data=None, profile_name=None, permutation_id=None):
        """Add a profile.

        Args:
            profile_data (dict, optional): The profile data.
            profile_name (str, optional): The name of the profile.
            permutation_id (int, optional): The index of the permutation id. Defaults to None.
        """
        profile_data = profile_data or {}
        if profile_name is not None:
            # profile_name is only used for profile migrations
            self.name = profile_name

        # get input permutations when a permutation_id is passed
        input_permutations = None
        if permutation_id is not None:
            try:
                input_permutations = self.permutations.input_dict(permutation_id)
            except Exception:
                # catch any error
                print(f'{c.Fore.RED}Invalid permutation id provided.')
                sys.exit(1)

        # this should not hit since tctest also check for duplicates
        if os.path.isfile(self.filename):  # pragma: no cover
            print(f'{c.Fore.RED}A profile with the name already exists.')
            sys.exit(1)

        profile = OrderedDict()
        profile['_comments_'] = []
        profile['environments'] = ['build']
        profile['stage'] = profile_data.get('stage', {'kvstore': {}})
        profile['configs'] = []  # add config here and remove later to ensure proper order
        profile['inputs'] = {}  # add inputs here and remove later to ensure proper order
        profile['trigger'] = {}
        profile['webhook_event'] = {
            'body': '',
            'headers': [],
            'method': 'GET',
            'query_params': [],
            'trigger_id': '',
        }
        profile['validation_criteria'] = {}
        profile['outputs'] = profile_data.get('outputs')
        profile['options'] = profile_data.get(
            'options', {'session': {'enabled': False, 'blur': []}},
        )

        if self.ij.runtime_level.lower() in ['triggerservice', 'webhooktriggerservice']:
            profile['configs'].extend(
                [
                    {
                        'trigger_id': str(randint(1000, 9999)),  # nosec
                        'config': profile_data.get(
                            'inputs',
                            {
                                'optional': self.ij.params_to_args(
                                    input_permutations=input_permutations,
                                    required=False,
                                    service_config=False,
                                ),
                                'required': self.ij.params_to_args(
                                    input_permutations=input_permutations,
                                    required=True,
                                    service_config=False,
                                ),
                            },
                        ),
                    }
                ]
            )
            del profile['inputs']  # inputs are for non service Apps
            del profile['validation_criteria']  # validation_criteria is for job Apps only
        elif self.ij.runtime_level.lower() in ['organization', 'playbook']:
            profile['exit_codes'] = profile_data.get('exit_codes', [0])
            profile['exit_message'] = None
            profile['inputs'].update(
                profile_data.get(
                    'inputs',
                    {
                        'optional': self.ij.params_to_args(
                            required=False, input_permutations=input_permutations
                        ),
                        'required': self.ij.params_to_args(
                            required=True, input_permutations=input_permutations
                        ),
                    },
                )
            )
            del profile['configs']  # inputs are for non service Apps
            del profile['trigger']  # trigger is for service Apps only
            del profile['webhook_event']  # webhook_event is for service Apps only

        if self.ij.runtime_level.lower() == 'organization':
            profile['stage']['threatconnect'] = {}
            profile['validation_criteria'] = profile_data.get('validation_criteria', {'percent': 5})

            del profile['outputs']  # outputs are not used in job Apps
        elif self.ij.runtime_level.lower() == 'playbook':
            del profile['validation_criteria']  # validation_criteria is for job Apps only
        elif self.ij.runtime_level.lower() == 'triggerservice':
            del profile['webhook_event']  # webhook_event is for webhooktriggerservice Apps only
        elif self.ij.runtime_level.lower() == 'webhooktriggerservice':
            del profile['trigger']  # trigger is for triggerservice Apps only

        # write the new profile to disk
        self.write(profile)

    def add_context(self, context):
        """Add a context to the context tracker for this profile.

        Args:
            context (str): The context (session_id) for this profile.
        """
        self._context_tracker.append(context)

    def clear_context(self, context):
        """Clear all context data in redis.

        Args:
            context (str): The context (session_id) to clear in KV store.
        """
        keys = self.redis_client.hkeys(context)
        if keys:
            return self.redis_client.hdel(context, *keys)
        return 0

    @property
    def contents(self):
        """Return mutable copy of profile JSON contents."""
        try:
            with open(self.filename) as fh:
                return json.load(fh, object_pairs_hook=OrderedDict)
        except (OSError, ValueError):
            print(f'{c.Fore.RED}Could not open/read profile {self.filename}.')

    @property
    def context_tracker(self):
        """Return the current context trackers for Service Apps."""
        if not self._context_tracker:
            if self.tcex_testing_context:
                self._context_tracker = json.loads(
                    self.redis_client.hget(self.tcex_testing_context, '_context_tracker') or '[]'
                )
        return self._context_tracker

    @property
    def data(self):
        """Return single instance copy of current profile."""
        if self._data is None:
            self._data = self.contents
            self.remove_comments(self._data)

        # APP-618 - used in custom test cases
        if self._data:
            self._data['name'] = self.name

        return self._data

    @data.setter
    def data(self, profile_data):
        """Set profile_data dict."""
        self._data = profile_data
        self.remove_comments(self._data)

    def delete(self):
        """Delete an existing profile."""
        raise NotImplementedError('The delete method is not currently implemented.')

    @property
    def directory(self):
        """Return fully qualified profile directory."""
        return os.path.join(self._app_path, 'tests', self.feature, 'profiles.d')

    @property
    def feature(self):
        """Return the current feature."""
        if self._feature is None:
            # when called in testing framework will get the feature from pytest env var.
            self._feature = self._test_case_data[0].split('/')[1].replace('/', '-')
        return self._feature

    @property
    def feature_directory(self):
        """Return fully qualified feature directory."""
        return os.path.join(self._app_path, 'tests', self.feature)

    @property
    def filename(self):
        """Return profile fully qualified filename."""
        return os.path.join(self.directory, f'{self.name}.json')
        # return os.path.join(self.directory, f'{self.name}.json')

    def init(self):
        """Return the Data (dict) from the current profile."""
        if self.data is None:
            self.log.error('Profile init failed; loaded profile data is None')

        # Now can initialize anything that needs initializing
        self.session_manager.init()  # initialize session recording/playback

    def merge_inputs(self):
        """Merge new inputs and remove undefined inputs."""
        if not self.pytest_args.get('merge_inputs'):
            return

        updated_params = []

        # handle non-layout and layout based App appropriately
        for profile_inputs, params in self.profile_inputs_params:
            profile_inputs_flattened = self._flatten_inputs(profile_inputs)

            inputs = {}
            merged_inputs = {
                'optional': {},
                'required': {},
            }
            for name, data in params.items():
                # inputs that are serviceConfig are not applicable for profiles
                if data.get('serviceConfig'):
                    continue

                # each non hidden input will be checked for permutations if the App has layout
                if not data.get('hidden'):
                    if not self.permutations.validate_input_variable(name, inputs):
                        continue

                # get the value from the current profile or use default value from install.json
                value = profile_inputs_flattened.get(name)
                if name not in profile_inputs_flattened and data.get('type').lower() != 'boolean':
                    # set the value to the default in the install.json file unless the type
                    # is boolean. changing a boolean value to True when not there will change
                    # the logic of the test case.
                    value = data.get('default', None)

                # get input type based on install.json required field
                input_type = 'optional'
                if data.get('required'):
                    input_type = 'required'

                # APP-87 - ensure boolean inputs don't have null values
                if data.get('type').lower() == 'boolean':
                    if not isinstance(value, bool):
                        value = False

                # update inputs for next permutation check
                inputs[name] = value

                # store merged/updated inputs for writing back to profile
                merged_inputs[input_type][name] = value

            # ADI-1376 - handle tcex default args (prevent removing)
            if profile_inputs.get('defaults'):
                merged_inputs['defaults'] = profile_inputs.get('defaults')
            updated_params.append(merged_inputs)

        # update the profile with merged config/inputs
        profile_data = self.contents
        if self.ij.runtime_level.lower() in ['triggerservice', 'webhooktriggerservice']:
            for index, config_item in enumerate(profile_data.get('configs', [])):
                config_item['config'] = updated_params[index]
        else:
            profile_data['inputs'] = updated_params[0]

        # write updated profile
        self.write(profile_data)

    def migrate(self):
        """Migration the profile to the latest schema"""
        migrate = Migrate(self)
        self.data = migrate.data

    @property
    def name(self):
        """Return partially parsed test case data."""
        if self._name is None:
            name_pattern = r'^test_[a-zA-Z0-9_]+\[(.+)\]$'
            self._name = re.search(name_pattern, self._test_case_data[-1]).group(1)
        return self._name

    @name.setter
    def name(self, name):
        """Set the profile name"""
        self._name = name

    def order_profile(self, profile_data):
        """Order the profile data properly."""
        comments = profile_data.pop('_comments_', None)
        environments = profile_data.pop('environments', None)
        exit_codes = profile_data.pop('exit_codes', None)
        exit_message = profile_data.pop('exit_message', None)
        configs = profile_data.pop('configs', None)
        inputs = profile_data.pop('inputs', None)
        options = profile_data.pop('options', None)
        outputs = profile_data.pop('outputs', None)
        stage = profile_data.pop('stage', None)
        trigger = profile_data.pop('trigger', None)
        validation_criteria = profile_data.pop('validation_criteria', None)
        webhook_event = profile_data.pop('webhook_event', None)

        profile = OrderedDict()
        if comments is not None:
            profile['_comments_'] = comments
        if environments is not None:
            profile['environments'] = environments
        if stage is not None:
            profile['stage'] = self._sorted(stage)
        if configs is not None:
            profile['configs'] = self._sorted(configs)
        if inputs is not None:
            profile['inputs'] = self._sorted(inputs)
        if trigger is not None:
            profile['trigger'] = self._sorted(trigger)
        if webhook_event is not None:
            profile['webhook_event'] = self._sorted(webhook_event)
        if validation_criteria is not None:
            profile['validation_criteria'] = validation_criteria
        if exit_message is not None:
            profile['exit_message'] = self._sorted(exit_message)
        if outputs is not None:
            profile['outputs'] = self._sorted(outputs)
        if options is not None:
            profile['options'] = self._sorted(options)
        if exit_codes is not None:
            profile['exit_codes'] = self._sorted(exit_codes)

        # add any additional fields not covered above
        for k, v in profile_data.items():
            profile[k] = v

        return profile

    @property
    def profile_inputs(self):
        """Return the appropriate inputs (config) for the current App type.

        Service App use config and others use inputs.

        "inputs": {
            "default": {}
            "optional": {}
            "required": {}
        }
        """
        if self.ij.runtime_level.lower() in ['triggerservice', 'webhooktriggerservice']:
            for config_data in self.configs:
                yield config_data.get('config')
        else:
            yield self.inputs

    @property
    def profile_inputs_params(self):
        """Return params for inputs."""
        # handle non-layout and layout based App appropriately
        for profile_inputs in self.profile_inputs:
            params = self.ij.params_dict  # params section of install.json build as dict
            if self.lj.has_layout:
                # using inputs from layout.json since they are required to be in order
                # (display field can only use inputs previously defined)
                params = {}
                for name in self.lj.params_dict:
                    # get data from install.json based on name
                    params[name] = self.ij.params_dict.get(name)

                # hidden fields will not be in layout.json so they need to be included manually
                params.update(self.ij.filter_params_dict(hidden=True))

            yield profile_inputs, params

    # TODO: BCS - move this
    @property
    def pytest_args(self):
        """Return dict of pytest config args."""
        if self._pytest_args is None:
            self._pytest_args = {}
            if self.pytestconfig:
                args = self.pytestconfig.option  # argparse.Namespace
                self._pytest_args = {
                    'merge_inputs': args.merge_inputs or False,
                    'merge_outputs': args.merge_outputs or False,
                    'replace_exit_message': args.replace_exit_message or False,
                    'replace_outputs': args.replace_outputs or False,
                    'record_session': args.record_session or False,
                    'ignore_session': args.ignore_session or False,
                }

        return self._pytest_args

    def remove_comments(self, data):
        """Iterate through data and remove any dict field with a value of "comments"

        Args:
            data (dict): The profile dictionary.
        """
        data = data or {}
        for _, v in list(data.items()):
            try:
                del data['_comments_']
            except KeyError:
                pass

            if isinstance(v, dict):
                self.remove_comments(v)

    @property
    def session(self):
        """Return a instance of the session manager."""
        if self._session is None:
            self._session = TcSession(
                self.env_store.getenv('/ninja/tc/tci/exchange_admin/api_access_id'),
                self.env_store.getenv('/ninja/tc/tci/exchange_admin/api_secret_key'),
                os.getenv('TC_API_PATH'),
            )
        return self._session

    @property
    def session_manager(self):
        """Return a instance of the session manager."""
        if not self._session_manager:
            self._session_manager = SessionManager(self)

        return self._session_manager

    @property
    def test_directory(self):
        """Return fully qualified test directory."""
        return os.path.join(self._app_path, 'tests')

    def update_exit_message(self):
        """Update validation rules from exit_message section of profile."""
        message_tc = ''
        if os.path.isfile(self.message_tc_filename):
            with open(self.message_tc_filename) as mh:
                message_tc = mh.read()

        profile_data = self.contents
        if (
            profile_data.get('exit_message') is None
            or isinstance(profile_data.get('exit_message'), str)
            or self.pytest_args.get('replace_exit_message')
        ):
            # update the profile
            profile_data['exit_message'] = {'expected_output': message_tc, 'op': 'eq'}

            self.write(profile_data, 'updating exit message')

    def update_outputs(self):
        """Update the validation rules for outputs section of a profile.

        By default this method will only update if the current value is null. If the
        flag --replace_outputs is passed to pytest (e.g., pytest --replace_outputs)
        the outputs will replaced regardless of their current value. If the flag
        --merge_outputs is passed to pytest (e.g., pytest --merge_outputs) any new
        outputs will be added and any outputs that are not longer valid will be
        removed.
        """
        if self.redis_client is None:
            # redis_client is only available for children of TestCasePlaybookCommon
            print(f'{c.Fore.RED}An instance of redis_client is not set.')
            sys.exit(1)

        outputs = {}
        trigger_id = None
        for context in self.context_tracker:
            # get all current keys in current context
            redis_data = self.redis_client.hgetall(context)
            trigger_id = self.redis_client.hget(context, '_trigger_id')

            # updated outputs with validation data
            self.update_outputs_variables(outputs, redis_data, trigger_id)

            # cleanup redis
            self.clear_context(context)

        if not self.outputs or self.pytest_args.get('replace_outputs'):
            # update profile if current profile is not or user specifies --replace_outputs
            profile_data = self.contents
            profile_data['outputs'] = outputs
            self.write(profile_data)
        elif self.pytest_args.get('merge_outputs'):
            if trigger_id is not None:
                # service Apps have a different structure with id: data
                merged_outputs = {}
                for id_, data in outputs.items():
                    merged_outputs[id_] = {}
                    for key in list(data):
                        if key in self.outputs.get(id_, {}):
                            # use current profile output value if exists
                            merged_outputs[id_][key] = self.outputs[id_][key]
                        else:
                            merged_outputs[id_][key] = outputs[id_][key]
            else:
                # update playbook App profile outputs
                merged_outputs = {}
                for key in list(outputs):
                    if key in self.outputs:
                        # use current profile output value if exists
                        merged_outputs[key] = self.outputs[key]
                    else:
                        merged_outputs[key] = outputs[key]

            profile_data = self.contents
            profile_data['outputs'] = merged_outputs
            self.write(profile_data)

    def update_outputs_variables(self, outputs, redis_data, trigger_id):
        """Return the outputs section of a profile.

        Args:
            outputs (dict): The dict to add outputs.
            redis_data (dict): The data from KV store for this profile.
            trigger_id (str): The current trigger_id (service Apps).
        """
        for variable in self.tc_playbook_out_variables:
            # TODO: investigate moving to output rules validator
            # APP-219 - check for "bad" output variable names
            if 'raw.json' in variable:
                self.log.data(
                    'validate',
                    'Suspect Value',
                    'Output variable matched a suspect value (raw.json).',
                    'warning',
                )

            # get data from redis for current context
            data = redis_data.get(variable.encode('utf-8'))

            # validate redis variables
            if data is None:
                if 1 not in self.exit_codes:
                    # TODO: add feature in testing framework to allow writing null and
                    #       then check if variables exist instead of null value.
                    # log error for missing output data if not a fail test case (exit code of 1)
                    self.log.data(
                        'validate',
                        'Missing Data',
                        f'possible missing KV Store data for variable {variable}',
                        'info',
                    )
            else:
                data = json.loads(data.decode('utf-8'))

            # validate validation variables
            validation_data = (self.outputs or {}).get(variable)
            if trigger_id is None and validation_data is None and self.outputs:
                self.log.error(f'[{self.name}] Missing validations rule: {variable}')

            # make business rules based on data type or content
            output_data = {'expected_output': data, 'op': 'eq'}
            if 1 not in self.exit_codes:
                output_data = self.rules.data(data)

            # get trigger id for service Apps
            if trigger_id is not None:
                if isinstance(trigger_id, bytes):
                    trigger_id = trigger_id.decode('utf-8')
                outputs.setdefault(trigger_id, {})
                outputs[trigger_id][variable] = output_data
            else:
                outputs[variable] = output_data

    def validate_inputs(self):
        """Validate required inputs.

        This method will also merge input if --merge_inputs is passed to pytest.
        """
        errors = []
        status = True

        # handle non-layout and layout based App appropriately
        for profile_inputs, params in self.profile_inputs_params:
            profile_inputs_flattened = self._flatten_inputs(profile_inputs)

            inputs = {}
            for name, data in params.items():
                if data.get('serviceConfig'):
                    # inputs that are serviceConfig are not applicable for profiles
                    continue

                if not data.get('hidden'):
                    # each non hidden input will be checked for permutations if the App has layout
                    if not self.permutations.validate_input_variable(name, inputs):
                        continue

                # get the value from the current profile or use default value from install.json
                value = profile_inputs_flattened.get(name)
                if name not in profile_inputs_flattened:
                    value = data.get('default', None)

                if data.get('required'):
                    if value in [None, '']:  # exclude 0 or False from check
                        # validation step
                        errors.append(f'- Missing/Invalid value for required arg ({name})')
                        status = False

                # APP-87 - ensure boolean inputs don't have null values
                if data.get('type').lower() == 'boolean':
                    if not isinstance(value, bool):
                        # validation step
                        errors.append(f'- Invalid value for boolean arg ({name})')
                        status = False

                # update inputs
                inputs[name] = value

        errors = '\n'.join(errors)  # convert error to string for assert message
        return status, f'\n{errors}'

    def write(self, json_data, reason=None):
        """Write updated profile file.

        Args:
            json_data (dict): The profile data.
            reason (str, default:None): The reason for the update.
        """
        # Permuted test cases set options to a true value, so disable writeback
        if self.test_options:
            return

        # order the profile data appropriately
        json_data = self.order_profile(json_data)

        if reason is not None:
            self.log.data(
                'profile', 'Profile Update', f'writing updated profile for {reason}', 'info'
            )

        with open(self.filename, 'w') as fh:
            json.dump(json_data, fh, indent=2, sort_keys=False)
            fh.write('\n')

    #
    # Properties
    #

    @property
    def args(self):
        """Return combined/flattened args."""
        return self.inputs_flattened

    @property
    def configs(self):
        """Return environments."""
        return list(self.data.get('configs', []))

    @property
    def environments(self):
        """Return environments."""
        return self.data.get('environments', ['build'])

    @property
    def exit_codes(self):
        """Return exit codes."""
        return self.data.get('exit_codes', [])

    @property
    def exit_message(self):
        """Return exit message dict."""
        return self.data.get('exit_message', {})

    @property
    def inputs(self):
        """Return inputs dict."""
        return self.data.get('inputs', {})

    @property
    def inputs_defaults(self):
        """Return required inputs dict."""
        return self.inputs.get('defaults', {})

    @property
    def inputs_flattened(self):
        """Return inputs dict."""
        return self._flatten_inputs(self.inputs)

    @property
    def inputs_optional(self):
        """Return required inputs dict."""
        return self.inputs.get('optional', {})

    @property
    def inputs_required(self):
        """Return required inputs dict."""
        return self.inputs.get('required', {})

    @property
    def message_tc_filename(self):
        """Return the fqpn for message_tc file relative to profile."""
        return os.path.join(
            self._default_args.get('tc_out_path'), self.feature, self._test_case_name, 'message.tc'
        )

    @property
    def options(self):
        """Return options dict."""
        if self.data.get('options') is None:
            self.data['options'] = {}
        return self.data.get('options')

    @property
    def owner(self):
        """Return the owner value."""
        return self.inputs_flattened.get('owner')

    @property
    def outputs(self):
        """Return outputs dict."""
        return self.data.get('outputs')

    @property
    def rargs(self):
        """Return combined/flattened args with value from staging data if required."""
        rargs = {}
        for arg, value in self.args.items():
            if re.match(self.utils.variable_match, value):
                # look for value in staging data
                if self.stage_kvstore.get(value) is not None:
                    value = self.stage_kvstore.get(value)
            rargs[arg] = value
        return rargs

    @property
    def stage(self):
        """Return stage dict."""
        if self.data.get('stage') is None:
            self.data['stage'] = {}
        return self.data.get('stage', {})

    @property
    def stage_kvstore(self):
        """Return stage kv store dict."""
        return self.stage.get('kvstore', {})

    @property
    def stage_threatconnect(self):
        """Return stage threatconnect dict."""
        return self.stage.get('threatconnect', {})

    @property
    def tc_in_path(self):
        """Return fqpn tc_in_path arg relative to profile."""
        if self.ij.runtime_level.lower() in [
            'apiservice',
            'triggerservice',
            'webhooktriggerservice',
        ]:
            tc_in_path = os.path.join(self._default_args.get('tc_in_path'), self.feature)
        else:
            tc_in_path = os.path.join(
                self._default_args.get('tc_in_path'), self.feature, self._test_case_name
            )
        return tc_in_path

    @property
    def tc_log_path(self):
        """Return fqpn tc_log_path arg relative to profile."""
        if self.ij.runtime_level.lower() in [
            'apiservice',
            'triggerservice',
            'webhooktriggerservice',
        ]:
            tc_log_path = os.path.join(self._default_args.get('tc_log_path'), self.feature)
        else:
            tc_log_path = os.path.join(
                self._default_args.get('tc_log_path'), self.feature, self._test_case_name
            )
        return tc_log_path

    @property
    def tc_out_path(self):
        """Return fqpn tc_out_path arg relative to profile."""
        if self.ij.runtime_level.lower() in [
            'apiservice',
            'triggerservice',
            'webhooktriggerservice',
        ]:
            tc_out_path = os.path.join(self._default_args.get('tc_out_path'), self.feature)
        else:
            tc_out_path = os.path.join(
                self._default_args.get('tc_out_path'), self.feature, self._test_case_name
            )
        return tc_out_path

    @property
    def tc_playbook_out_variables(self):
        """Return calculated output variables.

        * iterate over all inputs:
          * if input key has exposePlaybookKeyAs defined
          * if value a variable
            * lookup value in stage.kvstore data
            * for each key add to output variables
        """
        output_variables = self.ij.tc_playbook_out_variables
        if self.lj.has_layout:
            # if layout based App get valid outputs
            output_variables = self.ij.create_output_variables(
                self.permutations.outputs_by_inputs(self.inputs_flattened)
            )

        for arg, value in self.inputs_flattened.items():
            # get full input data from install.json
            input_data = self.ij.params_dict.get(arg, {})

            # check to see if it support dynamic output variables
            if 'exposePlaybookKeyAs' not in input_data:
                continue

            # get the output variable type from install.json input data
            variable_type = input_data.get('exposePlaybookKeyAs')

            # staged data for this dynamic input must be a KeyValueArray
            for data in self.stage_kvstore.get(value, []):
                # create a variable using key value
                variable = self.ij.create_variable(data.get('key'), variable_type, job_id=9876)
                output_variables.append(variable)

        # APP-77 - add _fired for service Apps
        if self.ij.runtime_level.lower() in ['triggerservice', 'webhooktriggerservice']:
            output_variables.append('#Trigger:9876:_fired!String')

        return output_variables

    @property
    def tc_temp_path(self):
        """Return fqpn tc_temp_path arg relative to profile."""
        if self.ij.runtime_level.lower() in [
            'apiservice',
            'triggerservice',
            'webhooktriggerservice',
        ]:
            tc_temp_path = os.path.join(self._default_args.get('tc_temp_path'), self.feature)
        else:
            tc_temp_path = os.path.join(
                self._default_args.get('tc_temp_path'), self.feature, self._test_case_name
            )
        return tc_temp_path

    @property
    def validation_criteria(self):
        """Return the validation_criteria value."""
        return self.data.get('validation_criteria', {})

    @property
    def webhook_event(self):
        """Return webhook event dict."""
        return self.data.get('webhook_event', {})

    @property
    def webhook_marshall_event(self):
        """Return webhook event dict."""
        return self.data.get('webhook_marshall_event', {})



"""App Decorators Module."""
# third-party
import wrapt


class WriteOutput:
    """Write the App output variables to Redis.

    This decorator will take the functions return value and write the data to Redis using the
    key and variable_type. An optional hard coded value can be passed, which will override the
    return value. If multiple value are provided for the same output variable there is an option
    to overwrite the previous value.

    This decorator is intended for very simple Apps. Using the `write_output()` method of the App
    template is the recommended way of writing output data.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        @WriteOutput(key='color', variable_type='String')
        def my_method(color):
            return color.lowercase()

    Args:
        key (str): The name of the playbook output variable.
        variable_type (str): The type for the playbook output variable.  Supported types are:
            String, Binary, KeyValue, TCEntity, TCEnhancedEntity, StringArray,
            BinaryArray, KeyValueArray, TCEntityArray, TCEnhancedEntityArray.
        default (str): If the method return is None use the provided value as a default.
        overwrite (bool): When True and more than one value is provided for the same variable
            the previous value will be overwritten.
    """

    def __init__(self, key, variable_type, default=None, overwrite=True):
        """Initialize Class properties"""
        self.key = key
        self.overwrite = overwrite
        self.default = default
        self.variable_type = variable_type

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def output(app, *args, **kwargs):
            """Call the function and store or append return value.

            Args:
                app (class): The instance of the App class "self".
            """
            data = wrapped(*args, **kwargs)
            if data is None and self.default is not None:
                data = self.default

            index = f'{self.key}-{self.variable_type}'
            if app.tcex.playbook.output_data.get(index) and not self.overwrite:
                # skip data since a previous value has already been written
                pass
            else:
                # store data returned by function call or default
                app.tcex.playbook.add_output(self.key, data, self.variable_type)
            return data

        return output(instance, *args, **kwargs)



"""App Decorators Module."""
# standard library
import inspect
from typing import List, Union

# third-party
import wrapt

# first-party
from tcex.validators import (
    ValidationError,
    equal_to,
    greater_than,
    greater_than_or_equal,
    in_range,
    less_than,
    less_than_or_equal,
    not_in,
    to_bool,
    to_float,
    to_int,
)


class IterateOnArg:
    """Iterate on values stored in ``self.args`` namespace.

    This decorator will iterate over all value of the supplied arg and return results. This feature
    is helpful when processing a single value (String) or array of values (StringArray). If the App
    was provided the arg ``self.args.colors`` with a value of ``['blue', 'green', 'magenta']``, then
    this decorator would call the function 3 times. Each call to the function would pass one value
    from the array. The return values are stored and returned all at once after the last value is
    processed.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # the args value of "colors" refers to ``self.args.colors``.
        @iterateOnArgs(arg='colors')
        def my_method(colors):
            return colors

        # ** OR **

        # the args value of "colors" refers to ``self.args.colors``.
        @iterateOnArg(arg='colors')
        def my_method(**kwargs):
            return kwargs.get('colors')

    Args:
        arg (str): The arg name from the App which contains the input. This input can be
            a Binary, BinaryArray, KeyValue, KeyValueArray, String, StringArray, TCEntity, or
            TCEntityArray.
        default (str, kwargs): Defaults to None. Default value to pass to method if arg
            value is None. Only supported for String or StringArray.
        fail_enabled (boolean|str, kwargs): Accepts a boolean or string value.  If a boolean value
            is provided that value will control enabling/disabling this feature. A string
            value should reference an item in the args namespace which resolves to a boolean.
            The value of this boolean will control enabling/disabling this feature.
        fail_on (list, kwargs): Defaults to None. Fail if data read from Redis is in list.
    """

    def __init__(self, arg, **kwargs):
        """Initialize Class properties"""
        self.arg = arg
        self.default = kwargs.get('default')
        self.fail_enabled = kwargs.get('fail_enabled', False)
        self.fail_msg = kwargs.get('fail_msg')
        self.fail_on = kwargs.get('fail_on', [])
        self.transforms: Union[List[callable], callable] = kwargs.get('transforms', [])
        self.validators: Union[List[callable], callable] = kwargs.get('validators', [])
        if self.fail_on:
            self.validators.insert(0, not_in(self.fail_on))

        self._init_validators(**kwargs)

    def _init_validators(self, **kwargs):
        validators_map = {
            'in_range': in_range,
            'equal_to': equal_to,
            'less_than': less_than,
            'less_than_or_equal': less_than_or_equal,
            'greater_than': greater_than,
            'greater_than_or_equal': greater_than_or_equal,
        }

        for k, v in filter(lambda item: item[0] in validators_map, kwargs.items()):
            validator = validators_map.get(k)
            if isinstance(v, list):
                self.validators.append(validator(*v))
            elif isinstance(v, dict):
                self.validators.append(validator(**v))
            else:
                self.validators.append(validator(v))

        transform_map = {
            'to_int': to_int,
            'to_float': to_float,
            'to_bool': to_bool,
        }

        for transform_name in filter(lambda t: t in transform_map, kwargs.keys()):
            transform = transform_map.get(transform_name)
            transform_args = kwargs.get(transform_name)
            if isinstance(transform_args, list):
                self.transforms.append(transform(*transform_args))
            elif isinstance(transform_args, dict):
                self.transforms.append(transform(**transform_args))
            else:
                self.transforms.append(transform())

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def loop(app, *args, **kwargs):
            """Iterate over data, calling the decorated function for each value.

            Args:
                app (class): The instance of the App class "self".
            """
            # retrieve the label for the current Arg
            label = app.tcex.ij.params_dict.get(self.arg, {}).get('label')

            # get the signature for the decorated method
            fn_signature = inspect.signature(wrapped, follow_wrapped=True).parameters

            # self.enable (e.g., True or 'fail_on_false') enables/disables this feature
            enabled = self.fail_enabled
            if not isinstance(self.fail_enabled, bool):
                enabled = getattr(app.args, self.fail_enabled)
                if not isinstance(enabled, bool):  # pragma: no cover
                    raise RuntimeError(
                        'The fail_enabled value must be a boolean or resolved to bool.'
                    )
                app.tcex.log.debug(f'Fail enabled is {enabled} ({self.fail_enabled}).')

            # retrieve data from Redis if variable and always return and array.
            results = []
            arg_data = app.tcex.playbook.read(getattr(app.args, self.arg))
            arg_type = app.tcex.playbook.variable_type(getattr(app.args, self.arg))
            if arg_data is None:
                arg_data = [None]
            elif not isinstance(arg_data, list):
                arg_data = [arg_data]

            _index = 0
            _array_length = len(arg_data)
            for ad in arg_data:

                # add "magic" args
                if '_index' in fn_signature:
                    kwargs['_index'] = _index
                if '_array_length' in fn_signature:
                    kwargs['_array_length'] = _array_length

                if ad is None and self.default is not None:
                    # set value passed to method to default if value is None.
                    ad = self.default
                    app.tcex.log.debug(
                        f'replacing null value with provided default value "{self.default}".'
                    )

                try:
                    for transform in self.transforms:
                        ad = transform(ad, self.arg, label)
                except ValidationError as v:
                    value_formatted = f'"{ad}"' if isinstance(ad, str) else str(ad)
                    message = f'Invalid value ({value_formatted}) found for "{label}": {v.message}'
                    app.tcex.log.error(message)
                    if self.fail_msg:
                        app.exit_message = self.fail_msg  # for test cases
                        app.tcex.exit(1, self.fail_msg)
                    else:
                        app.exit_message = message
                        app.tcex.exit(1, message)

                # check ad against fail_on_values
                if enabled:
                    try:
                        list([v(ad, self.arg, label) for v in self.validators])
                    except ValidationError as v:
                        value_formatted = f'"{ad}"' if isinstance(ad, str) else str(ad)
                        message = (
                            f'Invalid value ({value_formatted}) found for "{label}": {v.message}'
                        )
                        app.tcex.log.error(message)
                        if self.fail_msg:
                            app.exit_message = self.fail_msg  # for test cases
                            app.tcex.exit(1, self.fail_msg)
                        else:
                            app.exit_message = message
                            app.tcex.exit(1, message)

                # Add logging for debug/troubleshooting
                if (
                    arg_type not in ['Binary', 'BinaryArray']
                    and app.tcex.log.getEffectiveLevel() <= 10
                ):
                    log_string = str(ad)
                    if len(log_string) > 100:  # pragma: no cover
                        log_string = f'{log_string[:100]} ...'
                    app.tcex.log.debug(f'input value: {log_string}')

                # add results to kwargs
                kwargs[self.arg] = ad
                results.append(wrapped(*args, **kwargs))

                # increment index
                _index += 1
            return results

        return loop(instance, *args, **kwargs)



"""ThreatConnect TI Report"""
from ..group import Group


class Report(Group):
    """Unique API calls for Report API Endpoints

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        name (str): The name for this Group.
        file_name (str, kwargs): The name for the attached file for this Group.
        publish_date (str, kwargs): The publish datetime expression for this Group.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties."""

        super().__init__(
            tcex, sub_type='Report', api_entity='report', api_branch='reports', **kwargs
        )

    def file_content(self, file_content, update_if_exists=True):
        """Update  the file content.

        Args:
            file_content: The file_content to upload.
            update_if_exists:

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        self._data['fileContent'] = file_content
        return self.tc_requests.upload(
            self.api_type,
            self.api_branch,
            self.unique_id,
            file_content,
            update_if_exists=update_if_exists,
        )

    def file_name(self, file_name):
        """Update the file_name.

        Args:
            file_name:
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        self._data['fileName'] = file_name
        request = {'fileName': file_name}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def file_size(self, file_size):
        """Update the file_size.

        Args:
            file_size:
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        self._data['fileSize'] = file_size
        request = {'fileSize': file_size, 'fileName': self._data['fileName']}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def get_file_hash(self, hash_type='sha256'):
        """
        Getting the hash value of attached document
        Args:
            hash_type:

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.get_file_hash(
            self.api_type, self.api_branch, self.unique_id, hash_type=hash_type
        )

    def status(self, status):
        """Update the status

        Valid status:
        + Success
        + Awaiting Upload
        + In Progress
        + Failed

        Args:
            status: Success, Awaiting Upload, In Progress, or Failed
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        self._data['status'] = status
        request = {'status': status, 'fileName': self._data['fileName']}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def publish_date(self, publish_date):
        """Return Email to."""
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        publish_date = self._utils.datetime.format_datetime(
            publish_date, date_format='%Y-%m-%dT%H:%M:%SZ'
        )

        self._data['publishDate'] = publish_date
        request = {'publishDate': publish_date, 'fileName': self._data['fileName']}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def download(self):
        """Download the documents context.

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.download(self.api_type, self.api_branch, self.unique_id)
